import argparse
import builtins
import functools
import json
import logging
import multiprocessing
import multiprocessing.managers
import multiprocessing.pool
import numbers
import os
import queue
import socket
import subprocess
import warnings
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Optional, Tuple, Union, Any, Callable, Dict, Iterable

import requests


def is_number_between(
    value: numbers.Real, 
    lower: Optional[numbers.Real],
    upper: Optional[numbers.Real], 
    lower_close: bool = True,
    upper_close: bool = False
) -> bool:
    """Compares a value to upper and lower bounds, considering the bounds are closed or open.
    
    Args:
        value: A value of type numbers.Real, which represents the number to be checked.
        lower: An optional value of type numbers.Real, which represents the lower bound of the range. 
            It can be None if no lower bound is provided.
        upper: An optional value of type numbers.Real, which represents the upper bound of the range. 
            It can be None if no upper bound is provided.
        lower_close: A boolean value that indicates whether the lower bound should be considered closed or open. 
            If True, the lower bound is considered closed, meaning the value is included in the range. 
            If False, it is considered open, meaning the value is not included. Defaults to True.
        upper_close: A boolean value that indicates whether the upper bound should be considered closed or open. 
            If True, the upper bound is considered closed, meaning the value is included in the range. 
            If False, it is considered open, meaning the value is not included. Defaults to False.

    Notes:
        When `lower` is None, it can be understood as negative infinity. 
        When `upper` is None, it can be understood as positive infinity. 
        When the interval formed by `lower` and `upper` is not valid, the interval can be understood as an empty set.
    """
    assert isinstance(value, numbers.Real)
    assert lower is None or isinstance(lower, numbers.Real)
    assert upper is None or isinstance(upper, numbers.Real)
    assert isinstance(lower_close, bool)
    assert isinstance(upper_close, bool)
    
    if upper is not None:
        if upper_close:
            upper_result = value <= upper
        else:
            upper_result = value < upper
    else:
        upper_result = True
    
    if lower is not None:
        if lower_close:
            lower_result = lower <= value
        else:
            lower_result = lower < value
    else:
        lower_result = True
    return upper_result and lower_result


def all_of(iterable, pred):
    """Returns whether all elements in the iterable satisfy the predicate.

    Args:
        iterable (Iterable): An iterable to check.
        pred (callable): A predicate to apply to each element.

    Returns:
        bool: True if all elements satisfy the predicate, False otherwise.

    References:
        https://en.cppreference.com/w/cpp/algorithm/all_any_none_of
    """
    return all(pred(element) for element in iterable)


def any_of(iterable, pred):
    """Returns whether any element in the iterable satisfies the predicate.

    Args:
        iterable (Iterable): An iterable to check.
        pred (callable): A predicate to apply to each element.

    Returns:
        bool: True if any element satisfies the predicate, False otherwise.

    References:
        https://en.cppreference.com/w/cpp/algorithm/all_any_none_of
    """
    return any(pred(element) for element in iterable)


def none_of(iterable, pred):
    """Returns whether no elements in the iterable satisfy the predicate.

    Args:
        iterable (Iterable): An iterable to check.
        pred (callable): A predicate to apply to each element.

    Returns:
        bool: True if no elements satisfy the predicate, False otherwise.

    References:
        https://en.cppreference.com/w/cpp/algorithm/all_any_none_of
    """
    return not any(pred(element) for element in iterable)


def print_with_no(obj):
    if hasattr(obj, '__len__'):
        for k, item in enumerate(obj):
            print('[{}/{}] {}'.format(k+1, len(obj), item)) 
    elif hasattr(obj, '__iter__'):
        for k, item in enumerate(obj):
            print('[{}] {}'.format(k+1, item)) 
    else:
        print('[1] {}'.format(obj))
        
        
@dataclass
class TypeShapeSpec:
    name: str
    type: Optional[builtins.type] = None
    shape: Optional[Union[Tuple[int], int]] = None
    dtype: Optional[str] = None
    value: Optional[Union[int, bool, float]] = None
    
    def __repr__(self):
        type = self.type
        if self.dtype is not None:
            type = f'{type}[{self.dtype}]'
        repr_str = f"{self.name}: {type}, {self.shape}"
        if self.value is not None:
            repr_str = f'{repr_str}, {self.value}'
        return repr_str
    
    
def get_type_and_shape(x, name='x', max_seq_len=10) -> List[TypeShapeSpec]:
    """Determines the type and shape (sometimes dtype and value) of the given object x and returns a list of TypeShapeSpec objects.
  
    Args:  
        x (Any): The object to analyze.
        name (str, optional): The name of the object. Defaults to 'x'.
        max_seq_len (int, optional): The maximum sequence length to analyze for sequences. Defaults to 10.
  
    Returns:  
        List[TypeShapeSpec]: A list of TypeShapeSpec objects representing the type, shape, dtype, and value of x.
    """
    if (hasattr(x, 'shape') and not callable(x.shape) and
        hasattr(x, 'dtype') and not callable(x.dtype)):
        # e.g. torch.Tensor and numpy.ndarray
        results = [TypeShapeSpec(name, type(x), x.shape, str(x.dtype))]
    elif (hasattr(x, 'shape') and not callable(x.shape)):
        results = [TypeShapeSpec(name, type(x), x.shape)]
    elif isinstance(x, (tuple, list)):
        results = [TypeShapeSpec(name, type(x), len(x))]
        if max_seq_len is not None:
            seq_len = min(max_seq_len, len(x))
        for i in range(seq_len):
            results += get_type_and_shape(x[i], f"  {name}[{i}]", max_seq_len)
        if len(x) > seq_len:
            results += [TypeShapeSpec(f"  {name}[...]")]
    elif isinstance(x, dict):
        results = [TypeShapeSpec(name, type(x))]
        for k, v in x.items():
            results += get_type_and_shape(v, f"  {name}.{k}", max_seq_len)
    elif isinstance(x, str):
        results = [TypeShapeSpec(name, type(x), len(x))]
    elif isinstance(x, (int, bool, float)):
        results = [TypeShapeSpec(name, type(x), value=x)]
    else:
        results = [TypeShapeSpec(name, type(x))]
    return results


def get_file_line_count(filename, encoding='utf-8'):
    line_count = 0
    buffer_size = 1024 * 1024 * 8
    with open(filename, 'r', encoding=encoding) as f:
        while True:
            data = f.read(buffer_size)
            if not data:
                break
            line_count += data.count('\n')
    return line_count


def get_host_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip
    

def set_logger(filename, level=logging.INFO, logger_name=None, formatter=None, with_print=True):
    logger = logging.getLogger(logger_name) 
    logger.setLevel(level)
    
    if formatter is None:
        formatter = logging.Formatter('%(message)s')

    # Never mutate (insert/remove elements) the list you're currently iterating on. 
    # If you need, make a copy.
    for handler in logger.handlers[:]:
        if isinstance(handler, logging.FileHandler):
            logger.removeHandler(handler)
        # FileHandler is subclass of StreamHandler, so isinstance(handler,
        # logging.StreamHandler) is True even if handler is FileHandler.
        # if (type(handler) == logging.StreamHandler) and (handler.stream == sys.stderr):
        elif type(handler) == logging.StreamHandler:
            logger.removeHandler(handler)
            
    file_handler = logging.FileHandler(filename, encoding='utf-8')
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    if with_print:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    return logger


def print_arguments(args, sort_keys=True, indent=4):
    assert isinstance(args, argparse.Namespace)
    args = vars(args)
    print(json.dumps(args, indent=indent, sort_keys=sort_keys))


def save_arguments(filename, args, sort_keys=True, indent=4):
    assert isinstance(args, argparse.Namespace)
    args = vars(args)
    with open(filename, 'w') as f:
        json.dump(args, f, indent=indent, sort_keys=sort_keys)


class DownloadStatusCode(Enum):
    FILE_SIZE_TOO_LARGE = (-100, 'the size of file from url is too large')
    FILE_SIZE_TOO_SMALL = (-101, 'the size of file from url is too small')
    FILE_SIZE_IS_ZERO = (-102, 'the size of file from url is zero')
    URL_IS_NOT_IMAGE = (-103, 'URL is not an image')
    
    @property
    def code(self):
        return self.value[0]

    @property
    def message(self):
        return self.value[1]


class DownloadError(Exception):
    def __init__(self, status_code: DownloadStatusCode, extra_str: Optional[str] = None):
        self.status_code = status_code
        self.extra_str = extra_str
        
        self.name = status_code.name
        self.code = status_code.code
        if extra_str is None:
            self.message = status_code.message
        else:
            self.message = f'{status_code.message}: {extra_str}'
        Exception.__init__(self)

    def __repr__(self):
        return f'[{self.__class__.__name__} {self.code}] {self.message}'
    
    __str__ = __repr__

    
def download_image(image_url, min_filesize=0, max_filesize=100*1024*1024, 
                   params=None, **kwargs) -> bytes:
    """
    References:
        https://httpwg.org/specs/rfc9110.html#field.content-length
        https://requests.readthedocs.io/en/latest/user/advanced/#body-content-workflow
    """
    stream = kwargs.pop('stream', True)
    
    with requests.get(image_url, stream=stream, params=params, **kwargs) as response:
        response.raise_for_status()

        content_type = response.headers.get('content-type')
        if content_type is None:
            warnings.warn('No Content-Type!')
        else:
            if not content_type.startswith(('image/', 'application/octet-stream')):
                raise DownloadError(DownloadStatusCode.URL_IS_NOT_IMAGE)
        
        # when Transfer-Encoding == chunked, Content-Length does not exist.
        content_length = response.headers.get('content-length')
        if content_length is None:
            warnings.warn('No Content-Length!')
        else:
            content_length = int(content_length)
            if content_length > max_filesize:
                raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_LARGE)
            if content_length < min_filesize:
                raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_SMALL)
        
        filesize = 0
        chunks = []
        for chunk in response.iter_content(chunk_size=10*1024):
            chunks.append(chunk)
            filesize += len(chunk)
            if filesize > max_filesize:
                raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_LARGE)
        if filesize < min_filesize:
            raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_SMALL)
        image_bytes = b''.join(chunks)

    return image_bytes
    

def download_file(url, min_filesize=0, max_filesize=100*1024*1024, 
                  params=None, **kwargs) -> bytes:
    """
    References:
        https://httpwg.org/specs/rfc9110.html#field.content-length
        https://requests.readthedocs.io/en/latest/user/advanced/#body-content-workflow
    """
    stream = kwargs.pop('stream', True)
    
    with requests.get(url, stream=stream, params=params, **kwargs) as response:
        response.raise_for_status()

        # when Transfer-Encoding == chunked, Content-Length does not exist.
        content_length = response.headers.get('content-length')
        if content_length is None:
            warnings.warn('No Content-Length!')
        else:
            content_length = int(content_length)
            if content_length > max_filesize:
                raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_LARGE)
            if content_length < min_filesize:
                raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_SMALL)
        
        filesize = 0
        chunks = []
        for chunk in response.iter_content(chunk_size=10*1024):
            chunks.append(chunk)
            filesize += len(chunk)
            if filesize > max_filesize:
                raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_LARGE)
        if filesize < min_filesize:
            raise DownloadError(DownloadStatusCode.FILE_SIZE_TOO_SMALL)
        file_bytes = b''.join(chunks)

    return file_bytes
    

def get_git_repo_root(path: Optional[str] = None) -> Optional[str]:
    """Get the root directory of a Git repository.

    Args:
        path (Optional[str]): The path to the directory where the Git repository is located.
            If not provided, the current working directory will be used.

    Returns:
        Optional[str]: The absolute path of the top-level directory of the Git repository.
            If the Git repository is not found or an error occurs, None will be returned.
    """
    if path is None:
        path = os.getcwd()
        
    if not os.path.isdir(path):
        warnings.warn(f"The specified path '{path}' is not a valid directory.")
        return None
    
    try:
        output = subprocess.check_output(
            ['git', 'rev-parse', '--show-toplevel'],
            text=True,
            cwd=path
        ).strip()
        if not output:
            return None
        return os.path.normpath(os.path.realpath(output))
    except FileNotFoundError:
        warnings.warn("Git command not found. Ensure Git is installed and accessible in your PATH.")
        return None
    except subprocess.CalledProcessError as e:
        warnings.warn(f"Failed to execute Git command: {e}.")
        return None
    except Exception as e:
        warnings.warn(f"Error occurred while getting Git repo root: {e}")
        return None


def get_gpu_count() -> int:
    """Return the count of GPUs available on the system.

    This function uses the 'nvidia-smi' command to query the system for NVIDIA GPUs.
    It counts the occurrences of 'UUID' in the output to estimate the number of GPUs.

    Returns:
        int: The number of GPUs found, or 0 if an error occurs while executing the 'nvidia-smi' command.
    """
    try:
        output = subprocess.check_output(['nvidia-smi', '--list-gpus'], text=True)
        return output.count('UUID')
    except FileNotFoundError:
        warnings.warn("nvidia-smi command not found.")
        return 0
    except subprocess.CalledProcessError as e:
        warnings.warn(f"Error occurred while getting gpu count: {e}")
        return 0
    
    
def get_sw_slices(length: int, window_size: int, stride: Optional[int] = None,
                  drop_last: bool = False, adjust_last: bool = True) -> List[Tuple[int, int]]:
    """Generate a list of tuples representing sliding window slices of a sequence of given length.

    Args:
        length (int): The total length of the sequence to be sliced.
        window_size (int): The size of each sliding window.
        stride (Optional[int], optional): The step size (or stride) between the start of each window.
            If None, it defaults to the window_size. Defaults to None.
        drop_last (bool, optional): Whether to drop the last window if it is incomplete.
            Defaults to False.
        adjust_last (bool, optional): Whether to adjust the start index of the last window
            if it is incomplete and drop_last is False. This ensures the last window's size
            equals the window_size if possible. Defaults to True.

    Returns:
        List[Tuple[int, int]]: A list of tuples, where each tuple (start, end) represents
            a sliding window slice of the sequence. The end index is exclusive.

    Raises:
        TypeError: If `length`, `window_size`, or `stride` are not of the expected type.
        ValueError: If `length`, `stride`, or `window_size` are zero or negative.
    """
    if not isinstance(length, int):
        raise TypeError(f'`length` must be int, got {type(length)}')
    if not isinstance(window_size, int):
        raise TypeError(f'`window_size` must be int, got {type(window_size)}')
    if not isinstance(stride, int) and (stride is not None):
        raise TypeError(f'`stride` must be int or None, got {type(stride)}')
    if length <= 0:
        raise ValueError(f'`length` cannot be zero or negative, got {length}')
    if stride is not None and stride <= 0:
        raise ValueError(f'`stride` cannot be zero or negative, got {stride}')
    if window_size <= 0:
        raise ValueError(f'`window_size` cannot be zero or negative, got {window_size}')

    stride = stride or window_size

    slices = []
    max_index = min_index = 0
    while max_index < length:
        max_index = min_index + window_size
        if max_index > length:
            if not drop_last:
                max_index = length
                if adjust_last:
                    min_index = max(0, max_index - window_size)
                slices.append((min_index, max_index))
        else:
            slices.append((min_index, max_index))
        min_index = min_index + stride
    return slices


class QueuedTaskExecutor:
    def __init__(
        self, 
        original_func: Callable, 
        _queue: multiprocessing.managers.SyncManager.Queue
    ) -> None:
        self.original_func = original_func
        self._queue = _queue

    def __call__(self, item, *args, **kwargs):
        try:
            ret = self.original_func(item, *args, **kwargs)
        except Exception as e:
            self._queue.put((item, None, os.getpid()))
            raise e
        self._queue.put((item, ret, os.getpid()))
        return ret
    

def default_error_callback(e):
    print(f'Error: {e}')


def default_progress_callback(total: int, current: int, pid: int, item: Any, ret: Any):
    print(f'{datetime.now()} [{current}/{total}]<{pid}> {item}')
    
    
def map_async_ex(
    func: Callable, 
    iterable: Iterable, 
    extra_args: Tuple = (),
    extra_kwargs: Dict[str, Any] = {}, 
    chunksize: Optional[int] = None, 
    num_workers: Optional[int] = None,
    callback=None,
    error_callback: Optional[Callable[[BaseException], object]] = default_error_callback,
    progress_callback: Optional[Callable[[int, int, int, Any, Any], None]] = default_progress_callback
)-> multiprocessing.pool.MapResult:
    manager = multiprocessing.Manager()
    q = manager.Queue()
    func_with_queue = QueuedTaskExecutor(func, q)
    partial_func = functools.partial(func_with_queue, *extra_args, **extra_kwargs)
    
    if num_workers is None:
        num_workers = os.cpu_count()
    if chunksize is None:
        chunksize = (len(iterable) + num_workers - 1) // num_workers
    pool = multiprocessing.pool.Pool(processes=num_workers)
    async_results = pool.map_async(partial_func, iterable, chunksize=chunksize, 
                                   callback=callback, error_callback=error_callback)
    pool.close()

    num_completed = 0
    while True:
        try:
            if num_completed == len(iterable):
                break
            item, ret, pid = q.get_nowait()
            num_completed += 1
            if progress_callback is not None:
                progress_callback(len(iterable), num_completed, pid, item, ret)
        except queue.Empty:
            pass
    pool.join()
    return async_results



import numbers
import warnings

import khandy
import numpy as np


def crop_image(image, x_min, y_min, x_max, y_max, border_value=0):
    """Crop the given image at specified rectangular area.
    
    See Also:
        translate_image
        
    References:
        PIL.Image.crop
        tf.image.resize_image_with_crop_or_pad
    """
    assert khandy.is_numpy_image(image)
    assert isinstance(x_min, numbers.Integral) and isinstance(y_min, numbers.Integral)
    assert isinstance(x_max, numbers.Integral) and isinstance(y_max, numbers.Integral)
    assert (x_min < x_max) and (y_min < y_max)
    
    src_height, src_width = image.shape[:2]
    dst_height, dst_width = y_max - y_min, x_max - x_min
    channels = 1 if image.ndim == 2 else image.shape[2]
    
    if isinstance(border_value, (tuple, list)):
        assert len(border_value) == channels, \
            'Expected the num of elements in tuple equals the channels ' \
            'of input image. Found {} vs {}'.format(
                len(border_value), channels)
    else:
        border_value = (border_value,) * channels
    dst_image = khandy.create_solid_color_image(
        dst_width, dst_height, border_value, dtype=image.dtype)

    src_x_begin = max(x_min, 0)
    src_x_end   = min(x_max, src_width)
    dst_x_begin = src_x_begin - x_min
    dst_x_end   = src_x_end - x_min

    src_y_begin = max(y_min, 0)
    src_y_end   = min(y_max, src_height)
    dst_y_begin = src_y_begin - y_min
    dst_y_end   = src_y_end - y_min
    
    if (src_x_begin >= src_x_end) or (src_y_begin >= src_y_end):
        return dst_image
    dst_image[dst_y_begin: dst_y_end, dst_x_begin: dst_x_end, ...] = \
        image[src_y_begin: src_y_end, src_x_begin: src_x_end, ...]
    return dst_image
    

def crop_or_pad(image, x_min, y_min, x_max, y_max, border_value=0):
    warnings.warn('crop_or_pad will be deprecated, use crop_image instead!')
    return crop_image(image, x_min, y_min, x_max, y_max, border_value)


def crop_coords(boxes, image_width, image_height):
    """
    References:
        `mmcv.impad`
        `pad` in https://github.com/kpzhang93/MTCNN_face_detection_alignment
        `MtcnnDetector.pad` in https://github.com/AITTSMD/MTCNN-Tensorflow
    """
    x_mins = boxes[:, 0]
    y_mins = boxes[:, 1]
    x_maxs = boxes[:, 2]
    y_maxs = boxes[:, 3]
    dst_widths = x_maxs - x_mins
    dst_heights = y_maxs - y_mins
    
    src_x_begin = np.maximum(x_mins, 0)
    src_x_end   = np.minimum(x_maxs, image_width)
    dst_x_begin = src_x_begin - x_mins
    dst_x_end   = src_x_end - x_mins
    
    src_y_begin = np.maximum(y_mins, 0)
    src_y_end   = np.minimum(y_maxs, image_height)
    dst_y_begin = src_y_begin - y_mins
    dst_y_end   = src_y_end - y_mins

    coords = np.stack([dst_y_begin, dst_y_end, dst_x_begin, dst_x_end, 
                       src_y_begin, src_y_end, src_x_begin, src_x_end, 
                       dst_heights, dst_widths], axis=0)
    return coords
    

def crop_or_pad_coords(boxes, image_width, image_height):
    warnings.warn('crop_or_pad_coords will be deprecated, use crop_coords instead!')
    return crop_coords(boxes, image_width, image_height)


def center_crop_image(image, dst_width, dst_height, strict=True):
    """
    strict: 
        when True, raise error if src size is less than dst size. 
        when False, remain unchanged if src size is less than dst size, otherwise center crop.
    """
    assert khandy.is_numpy_image(image)
    assert isinstance(dst_width, numbers.Integral) and isinstance(dst_height, numbers.Integral)
    src_height, src_width = image.shape[:2]
    if strict:
        assert (src_height >= dst_height) and (src_width >= dst_width)

    crop_top = max((src_height - dst_height) // 2, 0)
    crop_left = max((src_width - dst_width) // 2, 0)
    cropped = image[crop_top: dst_height + crop_top, 
                    crop_left: dst_width + crop_left, ...]
    return cropped


def center_crop(image, dst_width, dst_height, strict=True):
    warnings.warn('center_crop will be deprecated, use center_crop_image instead!')
    return center_crop_image(image, dst_width, dst_height, strict)


def center_pad_image(image, dst_width, dst_height, strict=True):
    """
    strict: 
        when True, raise error if src size is greater than dst size. 
        when False, remain unchanged if src size is greater than dst size, otherwise center pad.
    """
    assert khandy.is_numpy_image(image)
    assert isinstance(dst_width, numbers.Integral) and isinstance(dst_height, numbers.Integral)
    
    src_height, src_width = image.shape[:2]
    if strict:
        assert (src_height <= dst_height) and (src_width <= dst_width)
    
    padding_x = max(dst_width - src_width, 0)
    padding_y = max(dst_height - src_height, 0)
    padding_top = padding_y // 2
    padding_left = padding_x // 2
    if image.ndim == 2:
        padding = ((padding_top, padding_y - padding_top), 
                   (padding_left, padding_x - padding_left))
    else:
        padding = ((padding_top, padding_y - padding_top), 
                   (padding_left, padding_x - padding_left), (0, 0))
    return np.pad(image, padding, 'constant')

    
def center_pad(image, dst_width, dst_height, strict=True):
    warnings.warn('center_pad will be deprecated, use center_pad_image instead!')
    return center_pad_image(image, dst_width, dst_height, strict)




import codecs
import os
import numbers
import warnings
from dataclasses import dataclass, field
from io import BytesIO

import cv2
import khandy
import numpy as np
from PIL import Image


@dataclass
class ImageSize:
    width: int
    height: int

    @property
    def cols(self) -> int:
        return self.width
    
    @property
    def rows(self) -> int:
        return self.height
    
    @property
    def area(self) -> int:
        return self.width * self.height
    
    @property
    def aspect_ratio(self) -> float:
        return self.width / self.height
    
    @property
    def ar(self) -> float:
        return self.width / self.height
    
    
def get_image_size(obj) -> ImageSize:
    if isinstance(obj, (str, os.PathLike)):
        with open(obj, 'rb') as f:
            image = Image.open(f)
            return ImageSize(width=image.width, height=image.height)
    elif isinstance(obj, Image.Image):
        return ImageSize(width=obj.width, height=obj.height)
    elif isinstance(obj, np.ndarray) and obj.ndim >= 2:
        return ImageSize(width=obj.shape[1], height=obj.shape[0])
    else:
        raise TypeError(f"Unexpected type {type(obj)}")
    

def _is_isobmff(buf):
    if len(buf) < 16 or buf[4:8] != b'ftyp':
        return False
    if len(buf) < int(codecs.encode(buf[0:4], 'hex'), 16):
        return False
    return True


def _get_ftyp(buf):
    ftyp_len = int(codecs.encode(buf[0:4], 'hex'), 16)
    major_brand = buf[8:12].decode(errors='ignore')
    minor_version = int(codecs.encode(buf[12:16], 'hex'), 16)
    compatible_brands = []
    for i in range(16, ftyp_len, 4):
        compatible_brands.append(buf[i:i+4].decode(errors='ignore'))
    return major_brand, minor_version, compatible_brands


def match_heic(buf):
    if _is_isobmff(buf):
        major_brand, minor_version, compatible_brands = _get_ftyp(buf)
        if major_brand == 'heic':
            return True
        if major_brand in ['mif1', 'msf1'] and 'heic' in compatible_brands:
            return True


def get_image_file_type(file_or_buffer):
    """Determine the image file type based on the file header.

    Args:
        file_or_buffer (Union[BinaryIO, bytes]): file_or_buffer: A file path or buffer object.

    Returns:
        Union[str, None]: A lowercase string representing the image file type ('bmp', 'gif', 'jpeg', 'png', 'tiff', 'webp', 'heic')
        or None if the file type is not recognized.

    References:
        https://bugs.python.org/issue28591
        https://github.com/h2non/filetype.py/blob/master/filetype/types/image.py
        https://github.com/kovidgoyal/calibre/blob/master/src/calibre/utils/imghdr.py
        https://peps.python.org/pep-0594/#imghdr
        https://github.com/sphinx-doc/sphinx/blob/a502e7523376e0344c1c9cc8a9d128143cd98b2d/sphinx/util/images.py
    """
    header = khandy.get_file_header(file_or_buffer, 32)

    # https://en.wikipedia.org/wiki/BMP_file_format#Bitmap_file_header
    if header.startswith(b'BM'):
        return 'bmp'

    # https://en.wikipedia.org/wiki/GIF#File_format
    if header.startswith((b'GIF87a', b'GIF89a')):
        return 'gif'

    # https://en.wikipedia.org/wiki/JPEG_File_Interchange_Format#File_format_structure
    if header.startswith(b'\xFF\xD8\xFF'):
        return 'jpeg'

    # https://en.wikipedia.org/wiki/PNG#File_header
    if header.startswith(b'\x89PNG\r\n\x1A\n'):
        return 'png'

    # https://en.wikipedia.org/wiki/TIFF#Byte_order
    if header.startswith((b'MM', b'II')):
        return 'tiff'

    # https://en.wikipedia.org/wiki/WebP#Technology
    if header.startswith(b'RIFF') and header[8:12] == b'WEBP':
        return 'webp'

    # https://en.wikipedia.org/wiki/High_Efficiency_Image_File_Format
    header = khandy.get_file_header(file_or_buffer, 8192)
    if match_heic(header):
        return 'heic'

    return None


def get_image_extension(file_or_buffer):
    warnings.warn('khandy.get_image_extension will be deprecated, use khandy.get_image_file_type instead!')
    return get_image_file_type(file_or_buffer)


def imread(file_or_buffer, flags=-1):
    """Improvement on cv2.imread, make it support filename including chinese character.
    
    Args:
        flags: same as OpenCV
    """
    try:
        if isinstance(file_or_buffer, (bytes, bytearray, memoryview)):
            return cv2.imdecode(np.frombuffer(file_or_buffer, dtype=np.uint8), flags)
        else:
            # support type: file or str or Path
            return cv2.imdecode(np.fromfile(file_or_buffer, dtype=np.uint8), flags)
    except Exception as e:
        warnings.warn(f'{e}')
        return None
    

def imread_cv(file_or_buffer, flags=-1):
    warnings.warn('khandy.imread_cv will be deprecated, use khandy.imread instead!')
    return imread(file_or_buffer, flags)


def imwrite(filename, image, params=None):
    """Improvement on cv2.imwrite, make it support filename including chinese character.
    """
    cv2.imencode(os.path.splitext(filename)[-1], image, params)[1].tofile(filename)


def imwrite_cv(filename, image, params=None):
    warnings.warn('khandy.imwrite_cv will be deprecated, use khandy.imwrite instead!')
    return imwrite(filename, image, params)


def imread_pil(file_or_buffer, to_mode=None):
    """Improvement on Image.open to avoid ResourceWarning.
    """
    try:
        if isinstance(file_or_buffer, bytes):
            buffer = BytesIO()
            buffer.write(file_or_buffer)
            buffer.seek(0)
            file_or_buffer = buffer

        if hasattr(file_or_buffer, 'read'):
            image = Image.open(file_or_buffer)
            if to_mode is not None:
                image = image.convert(to_mode)
        else:
            # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
            with open(file_or_buffer, 'rb') as f:
                image = Image.open(f)
                # If convert outside with statement, will raise "seek of closed file" as
                # https://github.com/microsoft/Swin-Transformer/issues/66
                if to_mode is not None:
                    image = image.convert(to_mode)
        return image
    except Exception as e:
        warnings.warn(f'{e}')
        return None
        
        
def imwrite_bytes(filename, image_bytes: bytes, raise_if_not_image: bool = True, 
                  use_adpative_ext: bool = True):
    """Write image bytes to file.
    
    Args:
        filename: str
            filename which image_bytes is written into.
        image_bytes: bytes
            image content to be written.
        raise_if_not_image: bool
            whether to raise if image_bytes is not image. Defaults to True.
        use_adpative_ext: bool
            whether to adaptively handle the file extension. Defaults to True.
    """
    extension = get_image_file_type(image_bytes)
    if extension is None and raise_if_not_image:
        raise Exception('image_bytes is not image!')
    if extension is not None:
        name_extension = khandy.get_path_extension(filename)
        if name_extension.lower() == '.jpg' and extension.lower() == 'jpeg':
            extension == 'jpg'
        if (extension.lower() != name_extension.lower()[1:]):
            if use_adpative_ext:
                filename = khandy.replace_path_extension(filename, extension)
            else:
                image = cv2.imdecode(np.frombuffer(image_bytes, np.uint8), -1)
                image_bytes = cv2.imencode(name_extension, image)[1]
    
    with open(filename, "wb") as f:
        f.write(image_bytes)
    return filename


def rescale_image(image: np.ndarray, rescale_factor='auto', dst_dtype=np.float32):
    """Rescale image by rescale_factor.

    Args:
        img (ndarray): Image to be rescaled.
        rescale_factor (str, int or float, *optional*, defaults to `'auto'`): 
            rescale the image by the specified scale factor. When is `'auto'`, 
            rescale the image to [0, 1).
        dtype (np.dtype, *optional*, defaults to `np.float32`):
            The dtype of the output image. Defaults to `np.float32`.

    Returns:
        ndarray: The rescaled image.
    """
    if rescale_factor == 'auto':
        if np.issubdtype(image.dtype, np.unsignedinteger):
            rescale_factor = 1. / np.iinfo(image.dtype).max
        else:
            raise TypeError(f'Only support uint dtype ndarray when `rescale_factor` is `auto`, got {image.dtype}')
    elif issubclass(rescale_factor, (int, float)):
        pass
    else:
        raise TypeError('rescale_factor must be "auto", int or float')
    image = image.astype(dst_dtype, copy=True)
    image *= rescale_factor
    image = image.astype(dst_dtype)
    return image


def normalize_image_value(image: np.ndarray, mean, std, rescale_factor=None):
    """Normalize image values using provided mean and standard deviation.  
  
    Args:
        image (np.ndarray): The input image to be normalized.
        mean (int, float, Sequence[int], Sequence[float], np.ndarray): The mean to be used for normalize.
        std (int, float, Sequence[int], Sequence[float], np.ndarray): The std to be used for normalize.
        rescale_factor (Union[int, float, str, None], optional): Rescale factor for mean and std.
            If set to 'auto', it will scale the mean and std based on the maximum value of the image dtype (only for unsigned integers).
            If int or float, the mean and std will be multiplied by the given factor.
            If None, no rescaling will be done. Defaults to None.
  
    Returns:
        np.ndarray: The normalized image with dtype np.float32.
  
    Raises:
        TypeError: If the `rescale_factor` is not one of the supported types (int, float, str, None).
        TypeError: If the image dtype is not unsigned integer when `rescale_factor` is set to 'auto'.
    """  
    dst_dtype = np.float32
    mean = np.array(mean, dtype=dst_dtype).flatten()
    std = np.array(std, dtype=dst_dtype).flatten()
    if rescale_factor == 'auto':
        if np.issubdtype(image.dtype, np.unsignedinteger):
            mean *= np.iinfo(image.dtype).max
            std *= np.iinfo(image.dtype).max
        else:
            raise TypeError(f'Only support uint dtype ndarray when `rescale_factor` is `auto`, got {image.dtype}')
    elif isinstance(rescale_factor, (int, float)):
        mean *= rescale_factor
        std *= rescale_factor
    image = image.astype(dst_dtype, copy=True)
    image -= mean
    image /= std
    return image


def normalize_image_dtype(image, keep_num_channels=False):
    """Normalize image dtype to uint8 (usually for visualization).
    
    Args:
        image : ndarray
            Input image.
        keep_num_channels : bool, optional
            If this is set to True, the result is an array which has 
            the same shape as input image, otherwise the result is 
            an array whose channels number is 3.
            
    Returns:
        out: ndarray
            Image whose dtype is np.uint8.
    """
    assert (image.ndim == 3 and image.shape[-1] in [1, 3]) or (image.ndim == 2)

    image = image.astype(np.float32)
    image = khandy.minmax_normalize(image, axis=None, copy=False)
    image = np.array(image * 255, dtype=np.uint8)
    
    if not keep_num_channels:
        if image.ndim == 2:
            image = np.expand_dims(image, -1)
        if image.shape[-1] == 1:
            image = np.tile(image, (1,1,3))
    return image
    
    
def normalize_image_channel(image, swap_rb=False):
    """Normalize image channel number and order to RGB or BGR.
    
    Args:
        image : ndarray
            Input image.
        swap_rb : bool, optional
            whether swap red and blue channel or not
            
    Returns:
        out: ndarray
            Image whose shape is (..., 3).
    """
    if image.ndim == 2:
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    elif image.ndim == 3:
        num_channels = image.shape[-1]
        if num_channels == 1:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
        elif num_channels == 3:
            if swap_rb:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        elif num_channels == 4:
            if swap_rb:
                image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGB)
            else:
                image = cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)
        else:
            raise ValueError(f'Unsupported image channel number, only support 1, 3 and 4, got {num_channels}!')
    else:
        raise ValueError(f'Unsupported image ndarray ndim, only support 2 and 3, got {image.ndim}!')
    return image


def normalize_image_shape(image, swap_rb=False):
    warnings.warn('khandy.normalize_image_shape will be deprecated, use khandy.normalize_image_channel instead!')
    return normalize_image_channel(image, swap_rb)


def stack_image_list(image_list, dtype=np.float32):
    """Join a sequence of image along a new axis before first axis.

    References:
        `im_list_to_blob` in `py-faster-rcnn-master/lib/utils/blob.py`
    """
    assert isinstance(image_list, (tuple, list))

    max_dimension = np.array([image.ndim for image in image_list]).max()
    assert max_dimension in [2, 3]
    max_shape = np.array([image.shape[:2] for image in image_list]).max(axis=0)
    
    num_channels = []
    for image in image_list:
        if image.ndim == 2:
            num_channels.append(1)
        else:
            num_channels.append(image.shape[-1])
    assert len(set(num_channels) - set([1])) in [0, 1]
    max_num_channels = np.max(num_channels)
    
    blob = np.empty((len(image_list), max_shape[0], max_shape[1], max_num_channels), dtype=dtype)
    for k, image in enumerate(image_list):
        blob[k, :image.shape[0], :image.shape[1], :] = np.atleast_3d(image).astype(dtype, copy=False)
    if max_dimension == 2:
        blob = np.squeeze(blob, axis=-1)
    return blob
    

def is_numpy_image(image):
    return isinstance(image, np.ndarray) and image.ndim in {2, 3}


def is_gray_image(image, tol=3):
    assert is_numpy_image(image)
    if image.ndim == 2:
        return True
    elif image.ndim == 3:
        num_channels = image.shape[-1]
        if num_channels == 1:
            return True
        elif num_channels == 3:
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            gray3 = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)
            mae = np.mean(cv2.absdiff(image, gray3))
            return mae <= tol
        elif num_channels == 4:
            rgb = cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)
            gray = cv2.cvtColor(rgb, cv2.COLOR_BGR2GRAY)
            gray3 = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)
            mae = np.mean(cv2.absdiff(rgb, gray3))
            return mae <= tol
        else:
            return False
    else:
        return False
        

def is_solid_color_image(image, tol=4):
    assert is_numpy_image(image)
    mean = np.array(cv2.mean(image)[:-1], dtype=np.float32)
    
    if image.ndim == 2:
        mae = np.mean(np.abs(image - mean[0]))
        return mae <= tol
    elif image.ndim == 3:
        num_channels = image.shape[-1]
        if num_channels == 1:
            mae = np.mean(np.abs(image - mean[0]))
            return mae <= tol
        elif num_channels == 3:
            mae = np.mean(np.abs(image - mean))
            return mae <= tol
        elif num_channels == 4:
            mae = np.mean(np.abs(image[:,:,:-1] - mean))
            return mae <= tol
        else:
            return False
    else:
        return False


def create_solid_color_image(image_width, image_height, color, dtype=None):
    if isinstance(color, numbers.Real):
        image = np.full((image_height, image_width), color, dtype=dtype)
    elif isinstance(color, (tuple, list)):
        if len(color) == 1:
            image = np.full((image_height, image_width), color[0], dtype=dtype)
        elif len(color) in (3, 4):
            image = np.full((1, 1, len(color)), color, dtype=dtype)
            image = cv2.copyMakeBorder(image, 0, image_height-1, 0, image_width-1, 
                                       cv2.BORDER_CONSTANT, value=color)
        else:
            color = np.asarray(color, dtype=dtype)
            image = np.empty((image_height, image_width, len(color)), dtype=dtype)
            image[:] = color
    else:
        raise TypeError(f'Invalid type {type(color)} for `color`.')
    return image



import warnings
from typing import Union

import numpy as np

try:
    import torch
    T = torch.Tensor
except:
    torch = None
    T = np.ndarray
KArray = Union[np.ndarray, T]


def is_torch_available() -> bool:
    return torch is not None


def is_torch_tensor(x) -> bool:
    return (torch is not None) and isinstance(x, torch.Tensor)


def import_torch():
    """
    References:
        https://github.com/ray-project/maze-raylit/blob/master/rllib/utils/framework.py
    """
    try:
        import torch
        return torch
    except ImportError as e:
        warnings.warn(f"PyTorch is not installed: {e}")
        return None



import cv2
import numpy as np


def get_similarity_transform(src_pts, dst_pts):
    """Get similarity transform matrix from src_pts to dst_pts
    
    Args:
        src_pts: Kx2 np.array
            source points matrix, each row is a pair of coordinates (x, y)
        dst_pts: Kx2 np.array
            destination points matrix, each row is a pair of coordinates (x, y)
            
    Returns:
        xform_matrix: 3x3 np.array
            transform matrix from src_pts to dst_pts
    """
    src_pts = np.asarray(src_pts)
    dst_pts = np.asarray(dst_pts)
    assert src_pts.shape == dst_pts.shape
    assert (src_pts.ndim == 2) and (src_pts.shape[-1] == 2)
    
    npts = src_pts.shape[0]
    src_x = src_pts[:, 0].reshape((-1, 1))
    src_y = src_pts[:, 1].reshape((-1, 1))
    tmp1 = np.hstack((src_x, -src_y, np.ones((npts, 1)), np.zeros((npts, 1))))
    tmp2 = np.hstack((src_y, src_x, np.zeros((npts, 1)), np.ones((npts, 1))))
    A = np.vstack((tmp1, tmp2))

    dst_x = dst_pts[:, 0].reshape((-1, 1))
    dst_y = dst_pts[:, 1].reshape((-1, 1))
    b = np.vstack((dst_x, dst_y))

    x = np.linalg.lstsq(A, b, rcond=-1)[0]
    x = np.squeeze(x)
    sc, ss, tx, ty = x[0], x[1], x[2], x[3]
    xform_matrix = np.array([
        [sc, -ss, tx],
        [ss,  sc, ty],
        [ 0,   0,  1]
    ])
    return xform_matrix
    
    
def align_and_crop(image, landmarks, std_landmarks, align_size, 
                   border_value=0, return_transform_matrix=False):
    landmarks = np.asarray(landmarks)
    std_landmarks = np.asarray(std_landmarks)
    xform_matrix = get_similarity_transform(landmarks, std_landmarks)

    landmarks_ex = np.pad(landmarks, ((0,0),(0,1)), mode='constant', constant_values=1)
    dst_landmarks = np.dot(landmarks_ex, xform_matrix[:2,:].T)
    dst_image = cv2.warpAffine(image, xform_matrix[:2,:], dsize=align_size, 
                               borderValue=border_value)
    if return_transform_matrix:
        return dst_image, dst_landmarks, xform_matrix
    else:
        return dst_image, dst_landmarks
        
        


import unittest
import khandy
import numpy as np


def random_boxes(x_min, y_min, x_max, y_max, num_boxes=1):
    random_xs = np.random.uniform(x_min, x_max, size=(num_boxes, 2))
    random_ys = np.random.uniform(y_min, y_max, size=(num_boxes, 2))
    x_mins, x_maxs = [], []
    for x in random_xs:
        x_mins.append(np.min(x))
        x_maxs.append(np.max(x))
    y_mins, y_maxs = [], []
    for y in random_ys:
        y_mins.append(np.min(y))
        y_maxs.append(np.max(y))
    return np.stack([x_mins, y_mins, x_maxs, y_maxs], axis=-1)


class TestDetObjects(unittest.TestCase):
    def setUp(self):
        self.det_objects = khandy.model.DetObjects(
            boxes=random_boxes(0, 0, 10, 10, num_boxes=10),
            classes=np.array([0, 1, 1, 3, 3, 5, 6, 7, 8, 9]),
            confs=np.linspace(0.1, 1.0, 10),
            class_names=['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
        )

    def test_filter_by_class_names(self):
        result1 = self.det_objects.filter_by_class_names(inplace=False)
        self.assertEqual(result1.class_names, ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'])

        result2 = self.det_objects.filter_by_class_names(ignored=['d', 'e'], inplace=False)
        self.assertEqual(result2.class_names, ['a', 'b', 'c', 'f', 'g', 'h', 'i', 'j'])

        result3 = self.det_objects.filter_by_class_names(interested=['a', 'b', 'c'], inplace=False)
        self.assertEqual(result3.class_names, ['a', 'b', 'c'])

        result4 = self.det_objects.filter_by_class_names(ignored=['d', 'e', 'x'], inplace=False)
        self.assertEqual(result4.class_names, ['a', 'b', 'c', 'f', 'g', 'h', 'i', 'j'])

        result5 = self.det_objects.filter_by_class_names(interested=['a', 'b', 'c', 'x'], inplace=False)
        self.assertEqual(result5.class_names, ['a', 'b', 'c'])

    def test_filter_by_class_indices(self):
        result1 = self.det_objects.filter_by_class_indices(inplace=False)
        self.assertTrue(np.array_equal(result1.classes.flatten(), [0, 1, 1, 3, 3, 5, 6, 7, 8, 9]))

        result2 = self.det_objects.filter_by_class_indices(ignored=[0, 1], inplace=False)
        self.assertTrue(np.array_equal(result2.classes.flatten(), [3, 3, 5, 6, 7, 8, 9]))

        result3 = self.det_objects.filter_by_class_indices(interested=[3, 4, 5], inplace=False)
        self.assertTrue(np.array_equal(result3.classes.flatten(), [3, 3, 5]))

        result4 = self.det_objects.filter_by_class_indices(ignored=[0, 1, 100], inplace=False)
        self.assertTrue(np.array_equal(result4.classes.flatten(), [3, 3, 5, 6, 7, 8, 9]))

        result5 = self.det_objects.filter_by_class_indices(interested=[3, 4, 5, 100], inplace=False)
        self.assertTrue(np.array_equal(result5.classes.flatten(), [3, 3, 5]))

    def _assert_equal(self, a, b):
        self.assertEqual(len(a), len(b))
        self.assertTrue(np.array_equal(a.boxes, b.boxes))
        self.assertTrue(np.array_equal(a.classes, b.classes))
        self.assertTrue(np.array_equal(a.confs, b.confs))
        self.assertEqual(a.class_names, b.class_names)

    def test_filter_misc(self):
        num_objects = 100
        det_objects = khandy.model.DetObjects(
            boxes=random_boxes(0, 0, 100, 100, num_boxes=num_objects),
            confs=np.random.uniform(0, 1, size=(num_objects,)),
        )
        boxes = khandy.Boxes(det_objects.boxes)
        self._assert_equal(det_objects.filter_by_conf(0.5), 
                           det_objects[det_objects.confs > 0.5])
        self._assert_equal(det_objects.filter_by_size(min_width=50), 
                           det_objects[boxes.widths[:, 0] >= 50])
        self._assert_equal(det_objects.filter_by_size(min_height=50), 
                           det_objects[boxes.heights[:, 0] >= 50])
        self._assert_equal(det_objects.filter_by_area(2500), 
                           det_objects[boxes.areas[:, 0] >= 2500])
        
    def test_filter_by_func(self):
        det_objects = khandy.model.DetObjects(
            boxes=np.array([[0, 0, 1, 1], [1, 1, 2, 2]]),
            confs=np.array([0.8, 0.9]),
            classes=np.array([0, 1]),
            class_names=["class_0", "class_1"]
        )
        
        def func(item: khandy.model.DetObjectItem) -> bool:
            return item.conf > 0.85
        
        filtered_objects = det_objects.filter_by_func(func)
        self.assertEqual(len(filtered_objects), 1)
        self.assertAlmostEqual(filtered_objects.confs[0], 0.9)
        self.assertEqual(len(det_objects), 2)
        self.assertAlmostEqual(det_objects.confs[0], 0.8)
        
        det_objects.filter_by_func(func, inplace=True)
        self.assertEqual(len(filtered_objects), 1)
        self.assertAlmostEqual(filtered_objects.confs[0], 0.9)
        self.assertEqual(len(det_objects), 1)
        self.assertAlmostEqual(det_objects.confs[0], 0.9)
        

class TestConcatDetObjects(unittest.TestCase):
    def setUp(self):
        self.boxes1 = np.random.randn(3, 4)
        self.boxes2 = np.random.randn(3, 4)
        self.boxes3 = np.random.randn(3, 4)
        self.det_objects1 = khandy.model.DetObjects(boxes=self.boxes1)
        self.det_objects2 = khandy.model.DetObjects(boxes=self.boxes2)
        self.det_objects3 = khandy.model.DetObjects(boxes=self.boxes3, extras=[1, 2, 5])
    
    def test_concat_all_fields(self):
        result = khandy.model.concat_det_objects([self.det_objects1, self.det_objects2])
        self.assertTrue(np.array_equal(result.boxes, np.concatenate([self.boxes1, self.boxes2])))
    
    def test_concat_common_fields(self):
        result = khandy.model.concat_det_objects([self.det_objects1, self.det_objects2, self.det_objects3], only_common_fields=True)
        self.assertTrue(np.array_equal(result.boxes, np.concatenate([self.boxes1, self.boxes2, self.boxes3])))
    
    def test_empty_list(self):
        print(khandy.model.concat_det_objects([]))

    def test_different_fields(self):
        with self.assertRaises(ValueError):
            khandy.model.concat_det_objects([self.det_objects1, self.det_objects2, self.det_objects3], only_common_fields=False)


if __name__ == '__main__':
    unittest.main()
    
    


import warnings
from dataclasses import dataclass
from typing import Literal, Tuple

import cv2
import numpy as np

import khandy

interp_codes = {
    'nearest': cv2.INTER_NEAREST,
    'bilinear': cv2.INTER_LINEAR,
    'bicubic': cv2.INTER_CUBIC,
    'area': cv2.INTER_AREA,
    'lanczos': cv2.INTER_LANCZOS4
}


def scale_image(image, x_scale, y_scale, return_scale=False, interpolation='bilinear'):
    """Scale image.
    
    Reference:
        mmcv.imrescale
    """
    assert khandy.is_numpy_image(image)
    src_height, src_width = image.shape[:2]
    dst_width = int(round(x_scale * src_width))
    dst_height = int(round(y_scale * src_height))
    
    resized_image = cv2.resize(image, (dst_width, dst_height), 
                               interpolation=interp_codes[interpolation])
    if not return_scale:
        return resized_image
    else:
        x_scale = dst_width / src_width
        y_scale = dst_height / src_height
        return resized_image, x_scale, y_scale


def resize_image(image, dst_width, dst_height, return_scale=False, interpolation='bilinear'):
    """Resize image to a given size.

    Args:
        image (ndarray): The input image.
        dst_width (int): Target width.
        dst_height (int): Target height.
        return_scale (bool): Whether to return `x_scale` and `y_scale`.
        interpolation (str): Interpolation method, accepted values are
            "nearest", "bilinear", "bicubic", "area", "lanczos".

    Returns:
        tuple or ndarray: (`resized_image`, `x_scale`, `y_scale`) or `resized_image`.
        
    Reference:
        mmcv.imresize
    """
    assert khandy.is_numpy_image(image)
    resized_image = cv2.resize(image, (dst_width, dst_height), 
                               interpolation=interp_codes[interpolation])
    if not return_scale:
        return resized_image
    else:
        src_height, src_width = image.shape[:2]
        x_scale = dst_width / src_width
        y_scale = dst_height / src_height
        return resized_image, x_scale, y_scale
    
    
def resize_image_short(image, dst_size, return_scale=False, interpolation='bilinear'):
    """Resize an image so that the length of shorter side is dst_size while 
    preserving the original aspect ratio.
    
    References:
        `resize_min` in `https://github.com/pjreddie/darknet/blob/master/src/image.c`
    """
    assert khandy.is_numpy_image(image)
    src_height, src_width = image.shape[:2]
    scale = max(dst_size / src_width, dst_size / src_height)
    dst_width = int(round(scale * src_width))
    dst_height = int(round(scale * src_height))
    
    resized_image = cv2.resize(image, (dst_width, dst_height), 
                               interpolation=interp_codes[interpolation])
    if not return_scale:
        return resized_image
    else:
        x_scale = dst_width / src_width
        y_scale = dst_height / src_height
        return resized_image, x_scale, y_scale
    
    
def resize_image_long(image, dst_size, return_scale=False, interpolation='bilinear'):
    """Resize an image so that the length of longer side is dst_size while 
    preserving the original aspect ratio.
    
    References:
        `resize_max` in `https://github.com/pjreddie/darknet/blob/master/src/image.c`
    """
    assert khandy.is_numpy_image(image)
    src_height, src_width = image.shape[:2]
    scale = min(dst_size / src_width, dst_size / src_height)
    dst_width = int(round(scale * src_width))
    dst_height = int(round(scale * src_height))
    
    resized_image = cv2.resize(image, (dst_width, dst_height), 
                               interpolation=interp_codes[interpolation])
    if not return_scale:
        return resized_image
    else:
        x_scale = dst_width / src_width
        y_scale = dst_height / src_height
        return resized_image, x_scale, y_scale
        
        
def resize_image_to_range(image, min_length, max_length, return_scale=False, interpolation='bilinear'):
    """Resizes an image so its dimensions are within the provided value.
    
    Rescale the shortest side of the image up to `min_length` pixels 
    while keeping the largest side below `max_length` pixels without 
    changing the aspect ratio. Often used in object detection (e.g. RCNN and SSH.)
    
    The output size can be described by two cases:
    1. If the image can be rescaled so its shortest side is equal to the
        `min_length` without the other side exceeding `max_length`, then do so.
    2. Otherwise, resize so the longest side is equal to `max_length`.
    
    Returns:
        resized_image: resized image so that
            min(dst_height, dst_width) == min_length or
            max(dst_height, dst_width) == max_length.
          
    References:
        `resize_to_range` in `models-master/research/object_detection/core/preprocessor.py`
        `prep_im_for_blob` in `py-faster-rcnn-master/lib/utils/blob.py`
        mmcv.imrescale
    """
    assert khandy.is_numpy_image(image)
    assert min_length < max_length
    src_height, src_width = image.shape[:2]
    
    min_side_length = min(src_width, src_height)
    max_side_length = max(src_width, src_height)
    scale = min_length / min_side_length
    if round(scale * max_side_length) > max_length:
        scale = max_length / max_side_length
    dst_width = int(round(scale * src_width))
    dst_height = int(round(scale * src_height))
    
    resized_image = cv2.resize(image, (dst_width, dst_height), 
                               interpolation=interp_codes[interpolation])
    if not return_scale:
        return resized_image
    else:
        x_scale = dst_width / src_width
        y_scale = dst_height / src_height
        return resized_image, x_scale, y_scale
        

@dataclass
class LetterBoxDetail:
    resized_w: int
    resized_h: int
    x_scale: float
    y_scale: float
    pad_top: int
    pad_left: int
    pad_bottom: int
    pad_right: int


LetterBoxLoc = Literal['top left', 'top center', 'top right',
                       'center left', 'center', 'center right', 
                       'bottom left', 'bottom center', 'bottom right']


def letterbox_image(image, dst_width, dst_height, border_value=0, interpolation='bilinear', 
                    loc: LetterBoxLoc = 'center') -> Tuple[np.ndarray, LetterBoxDetail]:
    """Resize an image while preserving its aspect ratio and pad it to match the desired dimensions.
  
    Args:  
        image (numpy.ndarray): The input image to be resized and padded.  
        dst_width (int): The desired width of the output image.  
        dst_height (int): The desired height of the output image.  
        border_value: The color value to use for padding. Defaults to 0 (black).  
        interpolation (str, optional): The interpolation method to use for resizing. Defaults to 'bilinear'.  
  
    Returns:  
        tuple: A tuple containing:  
            padded_image (numpy.ndarray): The resized and padded image.  
            lb_detail (LetterBoxDetail): An object containing details of the resizing and padding process.  

    References:
        letterbox_image` in `https://github.com/pjreddie/darknet/blob/master/src/image.c`
    """  
    assert khandy.is_numpy_image(image)
    src_height, src_width = image.shape[:2]
    scale = min(dst_width / src_width, dst_height / src_height)
    resized_w = int(round(scale * src_width))
    resized_h = int(round(scale * src_height))

    resized_image = cv2.resize(image, (resized_w, resized_h), interpolation=interp_codes[interpolation])
    
    height_diff = dst_height - resized_h
    width_diff = dst_width - resized_w
    pad_top, pad_left = {
        'top left': (0, 0),
        'top center': (0, width_diff // 2),
        'top right': (0, width_diff),
        'center left': (height_diff // 2, 0),
        'center': (height_diff // 2, width_diff // 2),
        'center right': (height_diff // 2, width_diff),
        'bottom left': (height_diff, 0),
        'bottom center': (height_diff, width_diff // 2),
        'bottom right': (height_diff, width_diff),
    }[loc]
    pad_bottom = height_diff - pad_top
    pad_right = width_diff - pad_left
    
    padded_image = cv2.copyMakeBorder(resized_image, pad_top, pad_bottom, pad_left, pad_right,
                                      cv2.BORDER_CONSTANT, value=border_value)
    
    lb_detail = LetterBoxDetail(
        resized_w=resized_w,
        resized_h=resized_h,
        x_scale=resized_w / src_width,
        y_scale=resized_h / src_height,
        pad_top=pad_top,
        pad_left=pad_left,
        pad_bottom=pad_bottom,
        pad_right=pad_right
    )
    return padded_image, lb_detail
        

def letterbox_resize_image(image, dst_width, dst_height, border_value=0,
                           return_scale=False, interpolation='bilinear'):
    warnings.warn('letterbox_resize_image will be deprecated, use letterbox_image instead!')
    dst_image, lb_detail = letterbox_image(image, dst_width, dst_height, border_value, interpolation)
    if not return_scale:
        return dst_image
    else:
        return dst_image, lb_detail.x_scale, lb_detail.pad_top, lb_detail.pad_left
    


import unittest
import khandy


class TestFliterDetectByLabel(unittest.TestCase):
    def setUp(self):
        self.detect_ir_objects = [
            khandy.label.DetectIrObject(label='label1', x_min=0, y_min=0, x_max=10, y_max=10), 
            khandy.label.DetectIrObject(label='label1', x_min=0, y_min=0, x_max=10, y_max=10), 
            khandy.label.DetectIrObject(label='label2', x_min=0, y_min=0, x_max=10, y_max=10)]  # Add more test objects

    def test_fliter_with_single_label(self):
        result = khandy.label.fliter_detect_by_label(self.detect_ir_objects, 'label1')
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].label, 'label1')
        self.assertEqual(result[1].label, 'label1')
        result = khandy.label.fliter_detect_by_label(self.detect_ir_objects, 'label2')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].label, 'label2')
        
    def test_fliter_with_multiple_labels(self):
        result = khandy.label.fliter_detect_by_label(self.detect_ir_objects, ['label1', 'label2'])
        self.assertEqual(len(result), 3)

    def test_fliter_with_callable(self):
        result = khandy.label.fliter_detect_by_label(self.detect_ir_objects, lambda x: x.startswith('label'))
        self.assertEqual(len(result), 3)

    def test_fliter_with_no_match(self):
        result = khandy.label.fliter_detect_by_label(self.detect_ir_objects, 'label3')
        self.assertEqual(len(result), 0)

    def test_fliter_with_empty_input(self):
        result = khandy.label.fliter_detect_by_label([], 'label1')
        self.assertEqual(result, [])


class TestDetectIrObject(unittest.TestCase):
    def test_initialization(self):
        obj =  khandy.label.DetectIrObject(label="test", x_min=1.0, y_min=1.0, x_max=2.0, y_max=2.0)
        self.assertEqual(obj.label, "test")
        self.assertEqual(obj.x_min, 1.0)
        self.assertEqual(obj.y_min, 1.0)
        self.assertEqual(obj.x_max, 2.0)
        self.assertEqual(obj.y_max, 2.0)

    def test_area(self):
        obj =  khandy.label.DetectIrObject(label="test", x_min=1.0, y_min=1.0, x_max=3.0, y_max=4.0)
        self.assertEqual(obj.area, 6.0)

    def test_swapping_values(self):
        with self.assertWarns(Warning):
            obj =  khandy.label.DetectIrObject(label="test", x_min=3.0, y_min=4.0, x_max=1.0, y_max=2.0)
        self.assertEqual(obj.x_min, 1.0)
        self.assertEqual(obj.x_max, 3.0)
        self.assertEqual(obj.y_min, 2.0)
        self.assertEqual(obj.y_max, 4.0)

 
if __name__ == '__main__':
    unittest.main()
    
    


import numpy as np


class FasterRcnnBoxCoder:
    """Faster RCNN box coder.
    
    Notes:
        boxes should be in cxcywh format.
    """
    
    def __init__(self, stddevs=None):
        """Constructor for FasterRcnnBoxCoder.
      
        Args:
          stddevs: List of 4 positive scalars to scale ty, tx, th and tw.
            If set to None, does not perform scaling. For Faster RCNN,
            the open-source implementation recommends using [0.1, 0.1, 0.2, 0.2].
        """
        if stddevs:
            assert len(stddevs) == 4
            for scalar in stddevs:
                assert scalar > 0
        self.stddevs = stddevs

    def encode(self, boxes, reference_boxes, copy=True):
        """Encode boxes with respect to reference boxes.
        """
        if copy:
            boxes = boxes.copy()
            
        boxes[..., 2:4] += 1e-8
        reference_boxes[..., 2:4] += 1e-8
        
        boxes[..., 0:2] -= reference_boxes[..., 0:2]
        boxes[..., 0:2] /= reference_boxes[..., 2:4]
        boxes[..., 2:4] /= reference_boxes[..., 2:4]
        boxes[..., 2:4] = np.log(boxes[..., 2:4], boxes[..., 2:4])
        if self.stddevs:
            boxes[..., 0:4] /= self.stddevs
        return boxes

    def decode(self, rel_boxes, reference_boxes, copy=True):
        """Decode relative codes to boxes.
        """
        if copy:
            rel_boxes = rel_boxes.copy()
            
        if self.stddevs:
            rel_boxes[..., 0:4] *= self.stddevs
        
        rel_boxes[..., 0:2] *= reference_boxes[..., 2:4]
        rel_boxes[..., 0:2] += reference_boxes[..., 0:2]
        rel_boxes[..., 2:4] = np.exp(rel_boxes[..., 2:4], rel_boxes[..., 2:4])
        rel_boxes[..., 2:4] *= reference_boxes[..., 2:4]
        return rel_boxes
    
    def decode_points(self, rel_points, reference_boxes, copy=True):
        """Decode relative codes to points.
        """
        if copy:
            rel_points = rel_points.copy()
        if self.stddevs:
            rel_points[..., 0::2] *= self.stddevs[0]
            rel_points[..., 1::2] *= self.stddevs[1]
        rel_points[..., 0::2] *= reference_boxes[..., 2:3]
        rel_points[..., 1::2] *= reference_boxes[..., 3:4]
        rel_points[..., 0::2] += reference_boxes[..., 0:1]
        rel_points[..., 1::2] += reference_boxes[..., 1:2]
        return rel_points



import numpy as np


def convert_xyxy_to_xywh(boxes, copy=True):
    """Convert [x_min, y_min, x_max, y_max] format to [x_min, y_min, width, height] format.
    """
    if copy:
        boxes = boxes.copy()
    boxes[..., 2:4] -= boxes[..., 0:2]
    return boxes


def convert_xywh_to_xyxy(boxes, copy=True):
    """Convert [x_min, y_min, width, height] format to [x_min, y_min, x_max, y_max] format.
    """
    if copy:
        boxes = boxes.copy()
    boxes[..., 2:4] += boxes[..., 0:2]
    return boxes


def convert_xywh_to_cxcywh(boxes, copy=True):
    """Convert [x_min, y_min, width, height] format to [cx, cy, width, height] format.
    """
    if copy:
        boxes = boxes.copy()
    boxes[..., 0:2] += boxes[..., 2:4] * 0.5
    return boxes
    
    
def convert_cxcywh_to_xywh(boxes, copy=True):
    """Convert [cx, cy, width, height] format to [x_min, y_min, width, height] format.
    """
    if copy:
        boxes = boxes.copy()
    boxes[..., 0:2] -= boxes[..., 2:4] * 0.5
    return boxes
    
    
def convert_xyxy_to_cxcywh(boxes, copy=True):
    """Convert [x_min, y_min, x_max, y_max] format to [cx, cy, width, height] format.
    """
    if copy:
        boxes = boxes.copy()
    boxes[..., 2:4] -= boxes[..., 0:2]
    boxes[..., 0:2] += boxes[..., 2:4] * 0.5
    return boxes


def convert_cxcywh_to_xyxy(boxes, copy=True):
    """Convert [cx, cy, width, height] format to [x_min, y_min, x_max, y_max] format.
    """
    if copy:
        boxes = boxes.copy()
    boxes[..., 0:2] -= boxes[..., 2:4] * 0.5
    boxes[..., 2:4] += boxes[..., 0:2]
    return boxes


def convert_boxes_format(boxes, in_fmt, out_fmt, copy=True):
    """Converts boxes from given in_fmt to out_fmt.

    Supported in_fmt and out_fmt are:
        'xyxy': boxes are represented via corners, x1, y1 being top left and x2, y2 being bottom right.
        'xywh' : boxes are represented via corner, width and height, x1, y2 being top left, w, h being width and height.
        'cxcywh' : boxes are represented via centre, width and height, cx, cy being center of box, w, h
            being width and height.

    Args:
        boxes: boxes which will be converted.
        in_fmt (str): Input format of given boxes. Supported formats are ['xyxy', 'xywh', 'cxcywh'].
        out_fmt (str): Output format of given boxes. Supported formats are ['xyxy', 'xywh', 'cxcywh']

    Returns:
        boxes: Boxes into converted format.

    References:
        torchvision.ops.box_convert
    """
    allowed_fmts = ("xyxy", "xywh", "cxcywh")
    if in_fmt not in allowed_fmts or out_fmt not in allowed_fmts:
        raise ValueError("Unsupported Bounding Box Conversions for given in_fmt and out_fmt")
    if copy:
        boxes = boxes.copy()
    if in_fmt == out_fmt:
        return boxes

    if (in_fmt, out_fmt) == ("xyxy", "xywh"):
        boxes = convert_xyxy_to_xywh(boxes, copy=False)
    elif (in_fmt, out_fmt) == ("xywh", "xyxy"):
        boxes = convert_xywh_to_xyxy(boxes, copy=False)
    elif (in_fmt, out_fmt) == ("xywh", "cxcywh"):
        boxes = convert_xywh_to_cxcywh(boxes, copy=False)
    elif (in_fmt, out_fmt) == ("cxcywh", "xywh"):
        boxes = convert_cxcywh_to_xywh(boxes, copy=False)
    elif (in_fmt, out_fmt) == ("xyxy", "cxcywh"):
        boxes = convert_xyxy_to_cxcywh(boxes, copy=False)
    elif (in_fmt, out_fmt) == ("cxcywh", "xyxy"):
        boxes = convert_cxcywh_to_xyxy(boxes, copy=False)
    return boxes
    


import numpy as np
from .boxes_utils import assert_and_normalize_shape


def scale_boxes(boxes, x_scale=1, y_scale=1, x_center=0, y_center=0, copy=True):
    """Scale boxes coordinates in x and y dimensions.
    
    Args:
        boxes: (N, 4+K)
        x_scale: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            scale factor in x dimension
        y_scale: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            scale factor in y dimension
        x_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        y_center: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        
    References:
        `core.box_list_ops.scale` in TensorFlow object detection API
        `utils.box_list_ops.scale` in TensorFlow object detection API
        `datasets.pipelines.Resize._resize_bboxes` in mmdetection
        `core.anchor.guided_anchor_target.calc_region` in mmdetection where comments may be misleading!
        `layers.mask_ops.scale_boxes` in detectron2
        `mmcv.bbox_scaling`
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)

    x_scale = np.asarray(x_scale, np.float32)
    y_scale = np.asarray(y_scale, np.float32)
    x_scale = assert_and_normalize_shape(x_scale, boxes.shape[0])
    y_scale = assert_and_normalize_shape(y_scale, boxes.shape[0])
    
    x_center = np.asarray(x_center, np.float32)
    y_center = np.asarray(y_center, np.float32)
    x_center = assert_and_normalize_shape(x_center, boxes.shape[0])
    y_center = assert_and_normalize_shape(y_center, boxes.shape[0])
    
    x_shift = 1 - x_scale
    y_shift = 1 - y_scale
    x_shift *= x_center
    y_shift *= y_center
    
    boxes[:, 0] *= x_scale
    boxes[:, 1] *= y_scale
    boxes[:, 2] *= x_scale
    boxes[:, 3] *= y_scale
    boxes[:, 0] += x_shift
    boxes[:, 1] += y_shift
    boxes[:, 2] += x_shift
    boxes[:, 3] += y_shift
    return boxes
    
    
def scale_boxes_wrt_centers(boxes, x_scale=1, y_scale=1, copy=True):
    """
    Args:
        boxes: (N, 4+K)
        x_scale: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            scale factor in x dimension
        y_scale: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            scale factor in y dimension
            
    References:
        `core.anchor.guided_anchor_target.calc_region` in mmdetection where comments may be misleading!
        `layers.mask_ops.scale_boxes` in detectron2
        `mmcv.bbox_scaling`
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)
    
    x_scale = np.asarray(x_scale, np.float32)
    y_scale = np.asarray(y_scale, np.float32)
    x_scale = assert_and_normalize_shape(x_scale, boxes.shape[0])
    y_scale = assert_and_normalize_shape(y_scale, boxes.shape[0])
    
    x_factor = (x_scale - 1) * 0.5
    y_factor = (y_scale - 1) * 0.5
    x_deltas = boxes[:, 2] - boxes[:, 0]
    y_deltas = boxes[:, 3] - boxes[:, 1]
    x_deltas *= x_factor
    y_deltas *= y_factor

    boxes[:, 0] -= x_deltas
    boxes[:, 1] -= y_deltas
    boxes[:, 2] += x_deltas
    boxes[:, 3] += y_deltas
    return boxes




import time
import logging
import numbers
import datetime
from typing import Callable, Any, Dict, Tuple, Optional

import numpy as np


def _to_timestamp(val, multiplier=1, rounded=False):
    if val is None:
        timestamp = time.time()
    elif isinstance(val, numbers.Real):
        timestamp = float(val)
    elif isinstance(val, time.struct_time):
        timestamp = time.mktime(val)
    elif isinstance(val, datetime.datetime):
        timestamp = val.timestamp()
    elif isinstance(val, datetime.date):
        dt = datetime.datetime.combine(val, datetime.time())
        timestamp = dt.timestamp()
    elif isinstance(val, str):
        try:
            # The full format looks like 'YYYY-MM-DD HH:MM:SS.mmmmmm'.
            dt = datetime.datetime.fromisoformat(val)
            timestamp = dt.timestamp()
        except:
            raise TypeError('when argument is str, it should conform to isoformat')
    else:
        raise TypeError('unsupported type!')
    timestamp = timestamp * multiplier
    if rounded:
        # The return value is an integer if ndigits is omitted or None.
        timestamp = round(timestamp)
    return timestamp


def get_timestamp(time_val=None, rounded=True):
    """timestamp in seconds.
    """
    return _to_timestamp(time_val, multiplier=1, rounded=rounded)


def get_timestamp_ms(time_val=None, rounded=True):
    """timestamp in milliseconds.
    """
    return _to_timestamp(time_val, multiplier=1000, rounded=rounded)


def get_timestamp_us(time_val=None, rounded=True):
    """timestamp in microseconds.
    """
    return _to_timestamp(time_val, multiplier=1000000, rounded=rounded)


def get_utc8now() -> datetime.datetime:
    """get current UTC+8 time or Beijing time
    """
    tz = datetime.timezone(datetime.timedelta(hours=8))
    utc8now = datetime.datetime.now(tz)
    return utc8now


class ContextTimer(object):
    """
    References:
        WithTimer in https://github.com/uber/ludwig/blob/master/ludwig/utils/time_utils.py
    """
    def __init__(self, name=None, use_log=False, quiet=False):
        self.use_log = use_log
        self.quiet = quiet
        if name is None:
            self.name = ''
        else:
            self.name = '{}, '.format(name.rstrip())
                
    def __enter__(self):
        self.start_time = time.time()
        if not self.quiet:
            self._print_or_log('{}{} starts'.format(self.name, self._now_time_str))
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.quiet:
            self._print_or_log('{}elapsed_time = {:.5}s'.format(self.name, self.get_eplased_time()))
            self._print_or_log('{}{} ends'.format(self.name, self._now_time_str))
            
    @property
    def _now_time_str(self):
        return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
    
    def _print_or_log(self, output_str):
        if self.use_log:
            logging.info(output_str)
        else:
            print(output_str)
            
    def get_eplased_time(self):
        return time.time() - self.start_time
        
    def enter(self):
        """Manually trigger enter"""
        self.__enter__()


def benchmark(
    func: Callable[..., Any], 
    args: tuple = (),
    kwargs: Dict[str, Any] = {},
    num_repeats: int = 100,
    num_repeats_burn_in: int = 2,
    display_interval: Optional[int] = None,
    display_desc: str = 'benchmark'
) -> Tuple[Any, Dict[str, float]]:
    """Run func several times to obtain the timing stats.
    
    References:
        tutorials/image/alexnet/alexnet_benchmark.py in tensorflow-models
    """
    assert isinstance(num_repeats, int) and num_repeats > 0
    assert isinstance(num_repeats_burn_in, int) and num_repeats_burn_in >= 0
    assert display_interval is None or isinstance(display_interval, int)

    duration_list = []
    for i in range(num_repeats + num_repeats_burn_in):
        start_time = time.time()
        output = func(*args, **kwargs)
        duration = time.time() - start_time
        if i >= num_repeats_burn_in:
            duration_list.append(duration)
            actual_step = i - num_repeats_burn_in + 1
            if display_interval is not None and actual_step % display_interval == 0:
                print(f'[{display_desc}][{actual_step}/{num_repeats}] {duration:.3f}s')
    
    summary = {
        'mean': np.mean(duration_list).item(),
        'stddev': np.std(duration_list).item(),
        'max': np.max(duration_list).item(),
        'min': np.min(duration_list).item(),
    }
    return output, summary




import numpy as np


def paired_intersection(boxes1, boxes2):
    """Compute paired intersection areas between boxes.
    Args:
        boxes1: a numpy array with shape [N, 4] holding N boxes
        boxes2: a numpy array with shape [N, 4] holding N boxes
        
    Returns:
        a numpy array with shape [N,] representing itemwise intersection area
        
    References:
        `core.box_list_ops.matched_intersection` in Tensorflow object detection API
        
    Notes:
        can called as itemwise_intersection, matched_intersection, aligned_intersection
    """
    max_x_mins = np.maximum(boxes1[:, 0], boxes2[:, 0])
    max_y_mins = np.maximum(boxes1[:, 1], boxes2[:, 1])
    min_x_maxs = np.minimum(boxes1[:, 2], boxes2[:, 2])
    min_y_maxs = np.minimum(boxes1[:, 3], boxes2[:, 3])
    intersect_widths = np.maximum(0, min_x_maxs - max_x_mins)
    intersect_heights = np.maximum(0, min_y_maxs - max_y_mins)
    return intersect_widths * intersect_heights
    
    
def pairwise_intersection(boxes1, boxes2):
    """Compute pairwise intersection areas between boxes.
    
    Args:
        boxes1: a numpy array with shape [N, 4] holding N boxes.
        boxes2: a numpy array with shape [M, 4] holding M boxes.
        
    Returns:
        a numpy array with shape [N, M] representing pairwise intersection area.
        
    References:
        `core.box_list_ops.intersection` in Tensorflow object detection API
        `utils.box_list_ops.intersection` in Tensorflow object detection API
    """
    if boxes1.shape[0] * boxes2.shape[0] == 0:
        return np.zeros((boxes1.shape[0], boxes2.shape[0]), dtype=boxes1.dtype)

    swap = False
    if boxes1.shape[0] > boxes2.shape[0]:
        boxes1, boxes2 = boxes2, boxes1
        swap = True
    intersect_areas = np.empty((boxes1.shape[0], boxes2.shape[0]), dtype=boxes1.dtype)

    for i in range(boxes1.shape[0]):
        max_x_mins = np.maximum(boxes1[i, 0], boxes2[:, 0])
        max_y_mins = np.maximum(boxes1[i, 1], boxes2[:, 1])
        min_x_maxs = np.minimum(boxes1[i, 2], boxes2[:, 2])
        min_y_maxs = np.minimum(boxes1[i, 3], boxes2[:, 3])
        intersect_widths = np.maximum(0, min_x_maxs - max_x_mins)
        intersect_heights = np.maximum(0, min_y_maxs - max_y_mins)
        intersect_areas[i, :] = intersect_widths * intersect_heights
    if swap:
        intersect_areas = intersect_areas.T
    return intersect_areas
    
    
def paired_overlap_ratio(boxes1, boxes2, ratio_type='iou'):
    """Compute paired overlap ratio between boxes.
    
    Args:
        boxes1: a numpy array with shape [N, 4] holding N boxes
        boxes2: a numpy array with shape [N, 4] holding N boxes
        ratio_type:
            iou: Intersection-over-union (iou).
            ioa: Intersection-over-area (ioa) between two boxes box1 and box2 is defined as
                their intersection area over box2's area. Note that ioa is not symmetric,
                that is, IOA(box1, box2) != IOA(box2, box1).
            min: Compute the ratio as the area of intersection between box1 and box2, 
                divided by the minimum area of the two bounding boxes.
                
    Returns:
        a numpy array with shape [N,] representing itemwise overlap ratio.
        
    References:
        `core.box_list_ops.matched_iou` in Tensorflow object detection API
        `structures.boxes.matched_boxlist_iou` in detectron2
        `mmdet.core.bbox.bbox_overlaps`, see https://mmdetection.readthedocs.io/en/v2.17.0/api.html#mmdet.core.bbox.bbox_overlaps
    """
    intersect_areas = paired_intersection(boxes1, boxes2)
    areas1 = (boxes1[:, 2] - boxes1[:, 0]) * (boxes1[:, 3] - boxes1[:, 1])
    areas2 = (boxes2[:, 2] - boxes2[:, 0]) * (boxes2[:, 3] - boxes2[:, 1])
    
    if ratio_type in ['union', 'iou', 'giou']:
        union_areas = areas1 - intersect_areas
        union_areas += areas2
        intersect_areas /= union_areas
    elif ratio_type == 'min':
        min_areas = np.minimum(areas1, areas2)
        intersect_areas /= min_areas
    elif ratio_type == 'ioa':
        intersect_areas /= areas2
    else:
        raise ValueError('Unsupported ratio_type. Got {}'.format(ratio_type))
        
    if ratio_type == 'giou':
        min_xy_mins = np.minimum(boxes1[:, 0:2], boxes2[:, 0:2])
        max_xy_mins = np.maximum(boxes1[:, 2:4], boxes2[:, 2:4])
        # mebb = minimum enclosing bounding boxes
        mebb_whs = np.maximum(0, max_xy_mins - min_xy_mins)
        mebb_areas = mebb_whs[:, 0] * mebb_whs[:, 1]
        union_areas -= mebb_areas
        union_areas /= mebb_areas
        intersect_areas += union_areas
    return intersect_areas


def pairwise_overlap_ratio(boxes1, boxes2, ratio_type='iou'):
    """Compute pairwise overlap ratio between boxes.
    
    Args:
        boxes1: a numpy array with shape [N, 4] holding N boxes
        boxes2: a numpy array with shape [M, 4] holding M boxes
        ratio_type:
            iou: Intersection-over-union (iou).
            ioa: Intersection-over-area (ioa) between two boxes box1 and box2 is defined as
                their intersection area over box2's area. Note that ioa is not symmetric,
                that is, IOA(box1, box2) != IOA(box2, box1).
            min: Compute the ratio as the area of intersection between box1 and box2, 
                divided by the minimum area of the two bounding boxes.
                
    Returns:
        a numpy array with shape [N, M] representing pairwise overlap ratio.
        
    References:
        `utils.np_box_ops.iou` in Tensorflow object detection API
        `utils.np_box_ops.ioa` in Tensorflow object detection API
        `utils.np_box_ops.giou` in Tensorflow object detection API
        `mmdet.core.bbox.bbox_overlaps`, see https://mmdetection.readthedocs.io/en/v2.17.0/api.html#mmdet.core.bbox.bbox_overlaps
        `torchvision.ops.box_iou`, see https://pytorch.org/vision/stable/ops.html#torchvision.ops.box_iou
        `torchvision.ops.generalized_box_iou`, see https://pytorch.org/vision/stable/ops.html#torchvision.ops.generalized_box_iou
        http://ww2.mathworks.cn/help/vision/ref/bboxoverlapratio.html
    """
    intersect_areas = pairwise_intersection(boxes1, boxes2)
    areas1 = (boxes1[:, 2] - boxes1[:, 0]) * (boxes1[:, 3] - boxes1[:, 1])
    areas2 = (boxes2[:, 2] - boxes2[:, 0]) * (boxes2[:, 3] - boxes2[:, 1])
    
    if ratio_type in ['union', 'iou', 'giou']:
        union_areas = np.expand_dims(areas1, axis=1) - intersect_areas
        union_areas += np.expand_dims(areas2, axis=0)
        intersect_areas /= union_areas
    elif ratio_type == 'min':
        min_areas = np.minimum(np.expand_dims(areas1, axis=1), np.expand_dims(areas2, axis=0))
        intersect_areas /= min_areas
    elif ratio_type == 'ioa':
        intersect_areas /= np.expand_dims(areas2, axis=0)
    else:
        raise ValueError('Unsupported ratio_type. Got {}'.format(ratio_type))
        
    if ratio_type == 'giou':
        min_xy_mins = np.minimum(boxes1[:, None, 0:2], boxes2[:, 0:2])
        max_xy_mins = np.maximum(boxes1[:, None, 2:4], boxes2[:, 2:4])
        # mebb = minimum enclosing bounding boxes
        mebb_whs = np.maximum(0, max_xy_mins - min_xy_mins)
        mebb_areas = mebb_whs[:, :, 0] * mebb_whs[:, :, 1]
        union_areas -= mebb_areas
        union_areas /= mebb_areas
        intersect_areas += union_areas
    return intersect_areas
    



from .boxes import *
from .boxes_clip import *
from .boxes_overlap import *
from .boxes_filter import *
from .boxes_convert import *
from .boxes_coder import *

from .boxes_transform_flip import *
from .boxes_transform_rotate import *
from .boxes_transform_scale import *
from .boxes_transform_translate import *
from .boxes_utils import *

from .boxes_and_indices import *



from typing import List, Optional

import numpy as np
from numpy.core.multiarray import normalize_axis_index


def sigmoid(x):
    return 1. / (1 + np.exp(-x))
    
    
def softmax(x: np.ndarray, axis=-1, valid_indices=None, copy=True):
    """Compute the softmax function.
      
    Args:  
        x (np.ndarray): The input array.
        axis (int, optional): The axis to perform the softmax operation. Defaults to -1.
        valid_indices (np.ndarray or None, optional): An array of valid indices to be used for fancy indexing.
            If None, all elements in x will be used. Defaults to None.
        copy (bool, optional): Whether to create a copy of the input array. Defaults to True.
          
    Returns:
        np.ndarray: The softmax output array.

    Referneces:
        sklearn.utils.extmath.softmax
        scipy.special.softmax
    """
    if copy:
        x = np.copy(x)
        
    if valid_indices is not None:
        # https://numpy.org/doc/stable/reference/generated/numpy.take.html
        # https://numpy.org/doc/stable/user/basics.copies.html
        # np.take returns a copy
        interested_x = np.take(x, valid_indices, axis=axis)
    else:
        interested_x = x
        
    max_val = np.max(interested_x, axis=axis, keepdims=True)
    interested_x -= max_val
    np.exp(interested_x, interested_x)
    sum_exp = np.sum(interested_x, axis=axis, keepdims=True)
    interested_x /= sum_exp
    
    # # If valid_indices are provided, place the softmax results back into the original array at the valid indices
    if valid_indices is not None:
        axis = normalize_axis_index(axis, x.ndim)
        x.fill(0)
        x[(slice(None),) * axis + (valid_indices,)] = interested_x
    else:
        x = interested_x
    return x


def log_sum_exp(x, axis=-1, keepdims=False):
    """
    References:
        numpy.logaddexp
        numpy.logaddexp2
        scipy.misc.logsumexp
    """
    max_val = np.max(x, axis=axis, keepdims=True)
    x -= max_val
    np.exp(x, x)
    sum_exp = np.sum(x, axis=axis, keepdims=keepdims)
    lse = np.log(sum_exp, sum_exp)
    if not keepdims:
        max_val = np.squeeze(max_val, axis=axis)
    return max_val + lse
    
    
def l2_normalize(x, axis=None, epsilon=1e-12, copy=True):
    """L2 normalize an array along an axis.
    
    Args:
        x : array_like of floats
            Input data.
        axis : None or int or tuple of ints, optional
            Axis or axes along which to operate.
        epsilon: float, optional
            A small value such as to avoid division by zero.
        copy : bool, optional
            Copy x or not.
    """
    if copy:
        x = np.copy(x)
    x /= np.maximum(np.linalg.norm(x, axis=axis, keepdims=True), epsilon)
    return x
    
    
def minmax_normalize(x, axis=None, epsilon=1e-12, copy=True):
    """minmax normalize an array along a given axis.
    
    Args:
        x : array_like of floats
            Input data.
        axis : None or int or tuple of ints, optional
            Axis or axes along which to operate.
        epsilon: float, optional
            A small value such as to avoid division by zero.
        copy : bool, optional
            Copy x or not.
    """
    if copy:
        x = np.copy(x)
    
    minval = np.min(x, axis=axis, keepdims=True)
    maxval = np.max(x, axis=axis, keepdims=True)
    maxval -= minval
    maxval = np.maximum(maxval, epsilon)
    
    x -= minval
    x /= maxval
    return x


def zscore_normalize(x, mean=None, std=None, axis=None, epsilon=1e-12, copy=True):
    """z-score normalize an array along a given axis.
    
    Args:
        x : array_like of floats
            Input data.
        mean:  array_like of floats, optional
            mean for z-score
        std: array_like of floats, optional
            std for z-score
        axis : None or int or tuple of ints, optional
            Axis or axes along which to operate.
        epsilon: float, optional
            A small value such as to avoid division by zero.
        copy : bool, optional
            Copy x or not.
    """
    if copy:
        x = np.copy(x)
    if mean is None:
        mean = np.mean(x, axis=axis, keepdims=True)
    if std is None:
        std = np.std(x, axis=axis, keepdims=True)
    mean = np.asarray(mean, dtype=x.dtype)
    std = np.asarray(std, dtype=x.dtype)
    std = np.maximum(std, epsilon)
    
    x -= mean
    x /= std
    return x


def get_order_of_magnitude(number):
    number = np.where(number == 0, 1, number)
    oom = np.floor(np.log10(np.abs(number)))
    return oom.astype(np.int32)
    
    
def top_k(x, k, axis=-1, largest=True, sorted=True):
    """Finds values and indices of the k largest/smallest 
    elements along a given axis.

    Args:
        x: numpy ndarray
            1-D or higher with given axis at least k.
        k: int
            Number of top elements to look for along the given axis.
        axis: int
            The axis to sort along.
        largest: bool
            Controls whether to return largest or smallest elements
        sorted: bool
            If true the resulting k elements will be sorted by the values.

    Returns:
        topk_values: 
            The k largest/smallest elements along the given axis.
        topk_indices: 
            The indices of the k largest/smallest elements along the given axis.
    """
    if axis is None:
        axis_size = x.size
    else:
        axis_size = x.shape[axis]
    assert 1 <= k <= axis_size

    x = np.asanyarray(x)
    if largest:
        index_array = np.argpartition(x, axis_size-k, axis=axis)
        topk_indices = np.take(index_array, -np.arange(k)-1, axis=axis)
    else:
        index_array = np.argpartition(x, k-1, axis=axis)
        topk_indices = np.take(index_array, np.arange(k), axis=axis)
    topk_values = np.take_along_axis(x, topk_indices, axis=axis)
    if sorted:
        sorted_indices_in_topk = np.argsort(topk_values, axis=axis)
        if largest:
            sorted_indices_in_topk = np.flip(sorted_indices_in_topk, axis=axis)
        sorted_topk_values = np.take_along_axis(
            topk_values, sorted_indices_in_topk, axis=axis)
        sorted_topk_indices = np.take_along_axis(
            topk_indices, sorted_indices_in_topk, axis=axis)
        return sorted_topk_values, sorted_topk_indices
    return topk_values, topk_indices
    

def sum_by_indices_list(x: np.ndarray, indices_list: List[List[int]], axis: int = -1) -> np.ndarray:
    """Sums of array elements according to a given list of index lists over a given axis.

    Args:
        x (np.ndarray): Input array to perform the sum operation.
        indices_list (List[List[int]]): List of index lists corresponding to the axis along which to sum.
        axis (int): Axis along which to perform the sum operation. Defaults to the last axis.

    Returns:
        np.ndarray: Output array after performing the sum operation along the specified axis.
            Has the same shape as the input array with the exception of the dimension along the axis 
            specified by indices_list, which has a length equal to the number of index lists in indices_list.
    """
    axis = normalize_axis_index(axis, x.ndim)
    new_shape = list(x.shape)
    new_shape[axis] = len(indices_list)
    dst = np.empty(new_shape, dtype=x.dtype)
    for new_index, old_indices in enumerate(indices_list):
        dest_dims = (slice(None),) * axis + (new_index,)
        dst[dest_dims] = np.sum(x.take(old_indices, axis), axis=axis, keepdims=False)
    return dst



import random
from collections import OrderedDict


def get_dict_first_item(dict_obj):
    for key in dict_obj:
        return key, dict_obj[key]


def sort_dict(dict_obj, key=None, reverse=False):
    return OrderedDict(sorted(dict_obj.items(), key=key, reverse=reverse))


def create_multidict(key_list, value_list):
    assert len(key_list) == len(value_list)
    multidict_obj = {}
    for key, value in zip(key_list, value_list):
        multidict_obj.setdefault(key, []).append(value)
    return multidict_obj


def convert_multidict_to_list(multidict_obj):
    key_list, value_list = [], []
    for key, value in multidict_obj.items():
        key_list += [key] * len(value)
        value_list += value
    return key_list, value_list


def convert_multidict_to_records(multidict_obj, key_map=None, raise_if_key_error=True):
    records = []
    if key_map is None:
        for key in multidict_obj:
            for value in multidict_obj[key]:
                records.append('{},{}'.format(value, key))
    else:
        for key in multidict_obj:
            if raise_if_key_error:
                mapped_key = key_map[key]
            else:
                mapped_key = key_map.get(key, key)
            for value in multidict_obj[key]:
                records.append('{},{}'.format(value, mapped_key))
    return records
    
    
def sample_multidict(multidict_obj, num_keys, num_per_key=None):
    num_keys = min(num_keys, len(multidict_obj))
    sub_keys = random.sample(list(multidict_obj), num_keys)
    if num_per_key is None:
        sub_mdict = {key: multidict_obj[key] for key in sub_keys}
    else:
        sub_mdict = {}
        for key in sub_keys:
            num_examples_inner = min(num_per_key, len(multidict_obj[key]))
            sub_mdict[key] = random.sample(multidict_obj[key], num_examples_inner)
    return sub_mdict
    
    
def split_multidict_on_key(multidict_obj, split_ratio, use_shuffle=False):
    """Split multidict_obj on its key.
    """
    assert isinstance(multidict_obj, dict)
    assert isinstance(split_ratio, (list, tuple))
    
    pdf = [k / float(sum(split_ratio)) for k in split_ratio]
    cdf = [sum(pdf[:k]) for k in range(len(pdf) + 1)]
    indices = [int(round(len(multidict_obj) * k)) for k in cdf]
    dict_keys = list(multidict_obj)
    if use_shuffle: 
        random.shuffle(dict_keys)
        
    be_split_list = []
    for i in range(len(split_ratio)):
        part_keys = dict_keys[indices[i]: indices[i + 1]]
        part_dict = dict([(key, multidict_obj[key]) for key in part_keys])
        be_split_list.append(part_dict)
    return be_split_list
    
    
def split_multidict_on_value(multidict_obj, split_ratio, use_shuffle=False):
    """Split multidict_obj on its value.
    """
    assert isinstance(multidict_obj, dict)
    assert isinstance(split_ratio, (list, tuple))
    
    pdf = [k / float(sum(split_ratio)) for k in split_ratio]
    cdf = [sum(pdf[:k]) for k in range(len(pdf) + 1)]
    be_split_list = [dict() for k in range(len(split_ratio))] 
    for key, value in multidict_obj.items():
        indices = [int(round(len(value) * k)) for k in cdf]
        cloned = value[:]
        if use_shuffle: 
            random.shuffle(cloned)
        for i in range(len(split_ratio)):
            be_split_list[i][key] = cloned[indices[i]: indices[i + 1]]
    return be_split_list
    
    
def get_multidict_info(multidict_obj, with_print=False, desc=None):
    num_list = [len(val) for val in multidict_obj.values()]
    num_keys = len(num_list)
    num_values = sum(num_list)
    max_values_per_key = max(num_list)
    min_values_per_key = min(num_list)
    if num_keys == 0:
        avg_values_per_key = 0
    else:
        avg_values_per_key = num_values / num_keys
    info = {
        'num_keys': num_keys,
        'num_values': num_values,
        'max_values_per_key': max_values_per_key,
        'min_values_per_key': min_values_per_key,
        'avg_values_per_key': avg_values_per_key,
    }
    if with_print:
        desc = desc or '<unknown>'
        print('{} key number:    {}'.format(desc, info['num_keys']))
        print('{} value number:    {}'.format(desc, info['num_values']))
        print('{} max number per-key: {}'.format(desc, info['max_values_per_key']))
        print('{} min number per-key: {}'.format(desc, info['min_values_per_key']))
        print('{} avg number per-key: {:.2f}'.format(desc, info['avg_values_per_key']))
    return info
    

def filter_multidict_by_number(multidict_obj, lower, upper=None):
    if upper is None:
        return {key: value for key, value in multidict_obj.items() 
                if lower <= len(value) }
    else:
        assert lower <= upper, 'lower must not be greater than upper'
        return {key: value for key, value in multidict_obj.items() 
                if lower <= len(value) <= upper }
        
        
def sort_multidict_by_number(multidict_obj, num_keys_to_keep=None, reverse=True):
    """
    Args:
        reverse: sort in ascending order when is True.
    """
    if num_keys_to_keep is None: 
        num_keys_to_keep = len(multidict_obj)
    else:
        num_keys_to_keep = min(num_keys_to_keep, len(multidict_obj))
    sorted_items = sorted(multidict_obj.items(), key=lambda x: len(x[1]), reverse=reverse)
    filtered_dict = OrderedDict()
    for i in range(num_keys_to_keep):
        filtered_dict[sorted_items[i][0]] = sorted_items[i][1]
    return filtered_dict

    
def merge_multidict(*mdicts):
    merged_multidict = {}
    for item in mdicts:
        for key, value in item.items():
            merged_multidict.setdefault(key, []).extend(value)
    return merged_multidict
    
    
def invert_multidict(multidict_obj):
    inverted_dict = {}
    for key, value in multidict_obj.items():
        for item in value:
            inverted_dict.setdefault(item, []).append(key)
    return inverted_dict
    
    


import os
import re
from setuptools import find_packages, setup


install_requires = ['numpy>=1.11.1', 'opencv-python', 'pillow', 'lxml', 'requests']


def get_version() -> str:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    version_file = os.path.join(current_dir,  "khandy/version.py")
    with open(version_file, encoding="utf-8") as f:
        version_match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M)
        if version_match:
            return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def get_long_description() -> str:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(current_dir, "README.md"), encoding="utf-8") as f:
        return f.read()
    

setup(
    name='khandy',
    version=get_version(),
    description='Handy Utilities for Computer Vision',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    keywords='computer vision',
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
    ],
    url='https://github.com/quarrying/khandy',
    author='quarryman',
    author_email='quarrying@qq.com',
    maintainer='quarryman',
    maintainer_email='quarrying@qq.com',
    license='MIT',
    install_requires=install_requires,
    zip_safe=False)



from .dict_utils import *
from .draw_utils import *
from .feature_utils import *
from .file_io_utils import *
from .fs_utils import *
from .hash_utils import *
from .seq_utils import *
from .misc import *
from .numpy_utils import *
from .optional_pkgs import *
from .text_utils import *
from .time_utils import *
from .version import *

from .boxes import *
from .image import *
from .points import *
from . import label
from . import model


import numpy as np
from .boxes_utils import assert_and_normalize_shape


def translate_boxes(boxes, x_shift=0, y_shift=0, copy=True):
    """translate boxes coordinates in x and y dimensions.
    
    Args:
        boxes: (N, 4+K)
        x_shift: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            shift in x dimension
        y_shift: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            shift in y dimension
        copy: bool
        
    References:
        `datasets.pipelines.RandomCrop` in mmdetection
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)
    
    x_shift = np.asarray(x_shift, np.float32)
    y_shift = np.asarray(y_shift, np.float32)

    x_shift = assert_and_normalize_shape(x_shift, boxes.shape[0])
    y_shift = assert_and_normalize_shape(y_shift, boxes.shape[0])
    
    boxes[:, 0] += x_shift
    boxes[:, 1] += y_shift
    boxes[:, 2] += x_shift
    boxes[:, 3] += y_shift
    return boxes
    
    
def adjust_boxes(boxes, x_min_shift, y_min_shift, x_max_shift, y_max_shift, copy=True):
    """
    Args:
        boxes: (N, 4+K)
        x_min_shift: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            shift (x_min, y_min) in x dimension
        y_min_shift: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            shift (x_min, y_min) in y dimension
        x_max_shift: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            shift (x_max, y_max) in x dimension
        y_max_shift: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
            shift (x_max, y_max) in y dimension
        copy: bool
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)

    x_min_shift = np.asarray(x_min_shift, np.float32)
    y_min_shift = np.asarray(y_min_shift, np.float32)
    x_max_shift = np.asarray(x_max_shift, np.float32)
    y_max_shift = np.asarray(y_max_shift, np.float32)

    x_min_shift = assert_and_normalize_shape(x_min_shift, boxes.shape[0])
    y_min_shift = assert_and_normalize_shape(y_min_shift, boxes.shape[0])
    x_max_shift = assert_and_normalize_shape(x_max_shift, boxes.shape[0])
    y_max_shift = assert_and_normalize_shape(y_max_shift, boxes.shape[0])
    
    boxes[:, 0] += x_min_shift
    boxes[:, 1] += y_min_shift
    boxes[:, 2] += x_max_shift
    boxes[:, 3] += y_max_shift
    return boxes
    
    
def inflate_or_deflate_boxes(boxes, width_delta=0, height_delta=0, copy=True):
    """
    Args:
        boxes: (N, 4+K)
        width_delta: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        height_delta: array-like whose shape is (), (1,), (N,), (1, 1) or (N, 1)
        copy: bool
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)

    width_delta = np.asarray(width_delta, np.float32)
    height_delta = np.asarray(height_delta, np.float32)

    width_delta = assert_and_normalize_shape(width_delta, boxes.shape[0])
    height_delta = assert_and_normalize_shape(height_delta, boxes.shape[0])
    
    half_width_delta = width_delta * 0.5
    half_height_delta = height_delta * 0.5
    boxes[:, 0] -= half_width_delta
    boxes[:, 1] -= half_height_delta
    boxes[:, 2] += half_width_delta
    boxes[:, 3] += half_height_delta
    return boxes
    

def inflate_boxes_to_square(boxes, copy=True):
    """Inflate boxes to square
    Args:
        boxes: (N, 4+K)
        copy: bool
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)

    widths = boxes[:, 2] - boxes[:, 0]
    heights = boxes[:, 3] - boxes[:, 1]
    max_side_lengths = np.maximum(widths, heights)
    
    width_deltas = np.subtract(max_side_lengths, widths, widths)
    height_deltas = np.subtract(max_side_lengths, heights, heights)
    width_deltas *= 0.5
    height_deltas *= 0.5
    boxes[:, 0] -= width_deltas
    boxes[:, 1] -= height_deltas
    boxes[:, 2] += width_deltas
    boxes[:, 3] += height_deltas
    return boxes
    

def deflate_boxes_to_square(boxes, copy=True):
    """Deflate boxes to square
    Args:
        boxes: (N, 4+K)
        copy: bool
    """
    boxes = np.array(boxes, dtype=np.float32, copy=copy)

    widths = boxes[:, 2] - boxes[:, 0]
    heights = boxes[:, 3] - boxes[:, 1]
    min_side_lengths = np.minimum(widths, heights)
    
    width_deltas = np.subtract(min_side_lengths, widths, widths)
    height_deltas = np.subtract(min_side_lengths, heights, heights)
    width_deltas *= 0.5
    height_deltas *= 0.5
    boxes[:, 0] -= width_deltas
    boxes[:, 1] -= height_deltas
    boxes[:, 2] += width_deltas
    boxes[:, 3] += height_deltas
    return boxes




import khandy
import numpy as np


def flip_image(image, direction='h', copy=True):
    """
    References:
        np.flipud, np.fliplr, np.flip
        cv2.flip
        tf.image.flip_up_down
        tf.image.flip_left_right
    """
    assert khandy.is_numpy_image(image)
    assert direction in ['x', 'h', 'horizontal',
                         'y', 'v', 'vertical', 
                         'o', 'b', 'both']
    if copy:
        image = image.copy()
    if direction in ['o', 'b', 'both', 'x', 'h', 'horizontal']:
        image = np.fliplr(image)
    if direction in ['o', 'b', 'both', 'y', 'v', 'vertical']:
        image = np.flipud(image)
    return image
    
    
def transpose_image(image, copy=True):
    """Transpose image.
    
    References:
        np.transpose
        cv2.transpose
        tf.image.transpose
    """
    assert khandy.is_numpy_image(image)
    if copy:
        image = image.copy()
    if image.ndim == 2:
        transpose_axes = (1, 0)
    else:
        transpose_axes = (1, 0, 2)
    image = np.transpose(image, transpose_axes)
    return image

    
def rot90_image(image, n=1, copy=True):
    """Rotate image counter-clockwise by 90 degrees.
    
    References:
        np.rot90
        cv2.rotate
        tf.image.rot90
    """
    assert khandy.is_numpy_image(image)
    if copy:
        image = image.copy()
    if image.ndim == 2:
        transpose_axes = (1, 0)
    else:
        transpose_axes = (1, 0, 2)
        
    n = n % 4
    if n == 0:
        return image[:]
    elif n == 1:
        image = np.transpose(image, transpose_axes)
        image = np.flipud(image)
    elif n == 2:
        image = np.fliplr(np.flipud(image))
    else:
        image = np.transpose(image, transpose_axes)
        image = np.fliplr(image)
    return image



from typing import Sequence, Tuple, Union

import numpy as np

import khandy


def sort_points(points: np.ndarray, axes: Union[int, Sequence[int]] = -1, 
                reverse: Union[bool, Sequence[bool]] = False) -> Tuple[np.ndarray, np.ndarray]:
    assert isinstance(points, np.ndarray) and points.ndim == 2

    axes = khandy.to_list(axes)
    if isinstance(reverse, bool):
        reverse = [reverse for _ in range(len(axes))]
    elif khandy.is_seq_of(reverse, bool):
        assert len(reverse) == len(axes)
    else:
        raise TypeError('reverse type should be bool or sequence of bool')

    maxs = np.max(points, axis=0)
    mins = np.min(points, axis=0)
    pos = np.zeros((len(points),), dtype=points.dtype)
    for k, (axis, reverse_on_axis) in enumerate(zip(axes, reverse)):
        if k != 0:
            pos *= maxs[axes[k-1]] - mins[axes[k-1]]
        if reverse_on_axis:
            pos += maxs[axis] - points[:, axis]
        else:
            pos += points[:, axis] - mins[axis]

    sorted_inds = np.argsort(pos)
    return sorted_inds, points[sorted_inds]




import cv2
import khandy
import numpy as np


def get_2d_rotation_matrix(angle, cx=0, cy=0, scale=1, 
                           degrees=True, dtype=np.float32):
    """
    References:
        `cv2.getRotationMatrix2D` in OpenCV
    """
    if degrees:
        angle = np.deg2rad(angle)
    c = scale * np.cos(angle)
    s = scale * np.sin(angle)

    tx = cx - cx * c + cy * s
    ty = cy - cx * s - cy * c
    return np.array([[ c, -s, tx],
                     [ s,  c, ty],
                     [ 0,  0, 1]], dtype=dtype)
    
    
def rotate_image(image, angle, scale=1.0, center=None, 
                 degrees=True, border_value=0, auto_bound=False):
    """Rotate an image.

    Args:
        image : ndarray
            Image to be rotated.
        angle : float
            Rotation angle in degrees, positive values mean clockwise rotation.
        center : tuple
            Center of the rotation in the source image, by default
            it is the center of the image.
        scale : float
            Isotropic scale factor.
        degrees : bool
        border_value : int
            Border value.
        auto_bound : bool
            Whether to adjust the image size to cover the whole rotated image.

    Returns:
        ndarray: The rotated image.
        
    References:
        mmcv.imrotate
    """
    assert khandy.is_numpy_image(image)
    image_height, image_width = image.shape[:2]
    if auto_bound:
        center = None
    if center is None:
        center = ((image_width - 1) * 0.5, (image_height - 1) * 0.5)
    assert isinstance(center, tuple)

    rotation_matrix = get_2d_rotation_matrix(angle, center[0], center[1], scale, degrees)
    if auto_bound:
        scale_cos = np.abs(rotation_matrix[0, 0])
        scale_sin = np.abs(rotation_matrix[0, 1])
        new_width = image_width * scale_cos + image_height * scale_sin
        new_height = image_width * scale_sin + image_height * scale_cos
        
        rotation_matrix[0, 2] += (new_width - image_width) * 0.5
        rotation_matrix[1, 2] += (new_height - image_height) * 0.5
        
        image_width = int(np.round(new_width))
        image_height = int(np.round(new_height))
    rotated = cv2.warpAffine(image, rotation_matrix[:2,:], (image_width, image_height), 
                             borderValue=border_value)
    return rotated



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

import khandy
torch = khandy.import_torch()


__all__ = ['ClassifierResultItem', 'ClassifierResults', 'BaseTopKClassifier',
           'BaseClassifier', 'ClassCollectionItem', 'BaseCollectionsClassifier',
           'Gallery', 'find_topk', 'find_topk_in_gallery']


def find_topk(inputs, indices_list: Optional[List[List[int]]] = None, 
              k: int = 3, do_softmax: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    if indices_list is not None:
        if do_softmax:
            valid_indices = list(set(sum(indices_list, [])))
            inputs = khandy.softmax(inputs, axis=-1, valid_indices=valid_indices)
        probs = khandy.sum_by_indices_list(inputs, indices_list, axis=-1)
    elif do_softmax:
        probs = khandy.softmax(inputs, axis=-1)
    else:
        probs = inputs
    topk_probs, topk_inds = khandy.top_k(probs, min(probs.shape[-1], k), axis=-1)
    return topk_probs, topk_inds


@dataclass
class ClassifierResultItem:
    conf: float
    class_index: int
    class_name: Optional[str] = None
    class_extra: Optional[Any] = None


class ClassifierResults(khandy.EqLenSequences):
    confs: khandy.KArray
    classes: khandy.KArray
    class_names: List[str]
    class_extras: List

    def __init__(self, confs: khandy.KArray, classes: Optional[khandy.KArray] = None, 
                 class_names: Optional[List[str]] = None, class_extras: Optional[List] = None):
        if confs.ndim == 2 and confs.shape[-1] == 1:
            confs = confs.flatten()
        if confs.ndim != 1:
            raise TypeError(f'Unsupported ndim for confs, got {confs.ndim}')
        
        if classes is None:
            if torch is not None and isinstance(confs, torch.Tensor):
                classes = torch.zeros((confs.shape[0],), dtype=torch.int32, device=confs.device)
            elif isinstance(confs, np.ndarray):
                classes = np.zeros((confs.shape[0],), dtype=np.int32)
            else:
                raise TypeError(f'Unsupported type for classes, got {type(classes)}')
        if classes.ndim == 2 and classes.shape[-1] == 1:
            classes = classes.flatten()
        if classes.ndim != 1:
            raise TypeError(f'Unsupported ndim for classes, got {classes.ndim}')
        
        assert class_names is None or khandy.is_list_of(class_names, str)
        assert class_extras is None or isinstance(class_extras, list)
        
        init_kwargs = {'confs': confs, 'classes': classes}
        if class_names is not None:
            init_kwargs['class_names'] = class_names
        if class_extras is not None:
            init_kwargs['class_extras'] = class_extras
        super().__init__(**init_kwargs)

    def __getitem__(self, key: Union[int, slice]) -> Union["ClassifierResults", ClassifierResultItem]:
        item = super().__getitem__(key)
        if type(key) == int:
            return ClassifierResultItem(
                conf=item.confs[0].item(),
                class_index=item.classes[0].item(),
                class_name=item.class_names[0] if 'class_names' in item else None,
                class_extra=item.class_extras[0] if 'class_extras' in item else None
            )
        return item
    
    
class BaseTopKClassifier(ABC):
    def __init__(
        self, 
        num_classes: int, 
        top_k: int = 3, 
        has_softmax: bool = False, 
        class_names: Optional[List[str]] = None,
        class_extras: Optional[List] = None
    ):
        self._num_classes = num_classes
        self._top_k = top_k
        self._has_softmax = has_softmax
        self._class_names = class_names
        self._class_extras = class_extras
        
    @property
    def num_classes(self) -> int:
        return self._num_classes

    @property
    def top_k(self) -> int:
        return self._top_k
    
    @top_k.setter
    def top_k(self, value: int):
        assert isinstance(value, int)
        if value <= 0:
            value = self.num_classes
        if value > self.num_classes:
            value = self.num_classes
        self._top_k = value
        
    @property
    def has_softmax(self) -> bool:
        return self._has_softmax

    @property
    def class_names(self) -> Optional[List[str]]:
        return self._class_names
    
    @class_names.setter
    def class_names(self, value: Optional[List[str]]):
        if value is not None:
            assert khandy.is_list_of(value, str) and len(value) == self.num_classes
        self._class_names = value
        
    @property
    def class_extras(self) -> Optional[List]:
        return self._class_extras
    
    @class_extras.setter
    def class_extras(self, value: Optional[List]):
        if value is not None:
            assert isinstance(value, list) and len(value) == self.num_classes
        self._class_extras = value
        
    @abstractmethod
    def forward(self, image: np.ndarray, **kwargs) -> np.ndarray:
        pass

    def __call__(self, image: np.ndarray, **kwargs) -> ClassifierResults:
        outputs = self.forward(image, **kwargs)
        topk_probs, topk_inds = find_topk(outputs, None, self.top_k, not self.has_softmax)
        results = ClassifierResults(confs=topk_probs[0], classes=topk_inds[0])
        if self.class_names is not None:
            results.class_names = [self.class_names[ind.item()] for ind in results.classes]
        if self.class_extras is not None:
            results.class_extras = [self.class_extras[ind.item()] for ind in results.classes]
        return results
    
    
class BaseClassifier(BaseTopKClassifier):
    def __init__(
        self, 
        num_classes: int, 
        has_softmax: bool = False, 
        class_names: Optional[List[str]] = None,
        class_extras: Optional[List] = None
    ):
        super().__init__(num_classes, 1, has_softmax, class_names, class_extras)

    def __call__(self, image: np.ndarray, **kwargs) -> ClassifierResultItem:
        return super().__call__(image, **kwargs)[0]


@dataclass
class ClassCollectionItem:
    indices: List[int]
    name: Optional[str] = None
    extra: Optional[Any] = None


def find_topk_in_collections(
        inputs, collections: Dict[str, List[ClassCollectionItem]], 
        k: int = 3, do_softmax: bool = False
    ) -> List[Dict[str, ClassifierResults]]:
    num_examples = len(inputs)
    results = [{} for _ in range(num_examples)]
    for collection_name, collection_val in collections.items(): # #collection
        indices_list = [item.indices for item in collection_val]
        topk_probs, topk_inds = find_topk(inputs, indices_list, k, do_softmax)
        for example_ind in range(num_examples): # batch_size
            one_results = ClassifierResults(confs=topk_probs[example_ind], classes=topk_inds[example_ind])
            class_names = []
            for ind in one_results.classes:
                class_name = collection_val[ind.item()].name
                if class_name is None:
                    class_name = f'{collection_name}#unnamed_class#{ind.item()}'
                class_names.append(class_name)
            one_results.class_names = class_names
            one_results.class_extras = [collection_val[ind.item()].extra for ind in one_results.classes]
            results[example_ind][collection_name] = one_results
    # (batch_size, #collection, k)
    return results


class BaseCollectionsClassifier(BaseTopKClassifier):
    def __init__(self, num_classes: int, 
                 collections: Dict[str, List[ClassCollectionItem]], 
                 top_k: int, has_softmax=False):
        super().__init__(num_classes, top_k, has_softmax)
        self._collections = collections

    @property
    def collections(self) -> Dict[str, List[ClassCollectionItem]]:
        return self._collections
    
    def __call__(self, image: np.ndarray, **kwargs) -> Dict[str, ClassifierResults]:
        outputs = self.forward(image, **kwargs)
        topk_results = find_topk_in_collections(outputs, self.collections, self.top_k, not self.has_softmax)
        return topk_results[0]


class Gallery:
    def __init__(self, features: np.ndarray, 
                 collections: Dict[str, List[ClassCollectionItem]], 
                 model_name: str):
        self._features = features
        self._collections = collections
        self._model_name = model_name
    
    @property
    def features(self) -> np.ndarray:
        return self._features
    
    @property
    def collections(self) -> Dict[str, List[ClassCollectionItem]]:
        return self._collections
    
    @property
    def model_name(self) -> str:
        return self._model_name
    
    @property
    def size(self) -> int:
        return len(self.features)
    
    @property
    def feature_dim(self) -> int:
        return self.features.shape[-1]


def find_topk_in_gallery(inputs, gallery: Gallery, k: int = 3) -> List[Dict[str, ClassifierResults]]:
    logits = np.dot(inputs, gallery.features.T)
    return find_topk_in_collections(logits, gallery.collections, k, do_softmax=True)




import unittest
import copy
import numpy as np
import khandy


class TestEqLenSequences(unittest.TestCase):
    def setUp(self):
        self.confs = np.arange(10) * 0.1
        self.class_names = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        self.data = khandy.EqLenSequences(
            confs=self.confs,
            class_names=self.class_names
        )

    def _assert_equal(self, a, b):
        self.assertEqual(len(a), len(b))
        self.assertTrue(np.array_equal(a.confs, b.confs))
        self.assertEqual(a.class_names, b.class_names)
        
    def test_getitem(self):
        self._assert_equal(self.data[0], khandy.EqLenSequences(
            confs=self.confs[0:1], class_names=self.class_names[0:1]))
        self._assert_equal(self.data[-1], khandy.EqLenSequences(
            confs=self.confs[-1:], class_names=self.class_names[-1:]))
        self._assert_equal(self.data[1:3], khandy.EqLenSequences(
            confs=self.confs[1:3], class_names=self.class_names[1:3]))
        self._assert_equal(self.data[[1, 2, 3]], khandy.EqLenSequences(
            confs=self.confs[[1, 2, 3]], class_names=self.class_names[1:4]))
        self._assert_equal(self.data[np.array([1, 2, 3])], khandy.EqLenSequences(
            confs=self.confs[np.array([1, 2, 3])], class_names=self.class_names[1:4]))
        
        mask = self.data.confs > 0.5
        self._assert_equal(self.data[mask], khandy.EqLenSequences(
            confs=self.confs[mask], class_names=['6', '7', '8', '9']))

    def _test_filter(self, data: khandy.EqLenSequences, indices, inplace=False):
        if inplace:
            stub = copy.deepcopy(data)
            filtered = stub.filter(indices, inplace=True)
            self._assert_equal(stub, self.data[indices])
            self._assert_equal(filtered, self.data[indices])
        else:
            stub = data
            filtered = stub.filter(indices, inplace=False)
            self._assert_equal(stub, self.data)
            self._assert_equal(filtered, self.data[indices])
        
    def test_filter(self):
        for inplace in [True, False]:
            self._test_filter(self.data, 0, inplace=inplace)
            self._test_filter(self.data, -1, inplace=inplace)
            self._test_filter(self.data, slice(1,3), inplace=inplace)
            self._test_filter(self.data, [1,2,3], inplace=inplace)
            self._test_filter(self.data, np.array([1, 2, 3]), inplace=inplace)
            self._test_filter(self.data, self.confs > 0.5, inplace=inplace)
        

if __name__ == '__main__':
    unittest.main()



import cv2
import khandy
import numpy as np


def _convert_bool_matrix_to_int(bool_mat):
    hash_val = int(0)
    for item in bool_mat.flatten():
        hash_val <<= 1
        hash_val |= int(item)
    return hash_val
    
    
def calc_image_ahash(image):
    """Average Hashing

    References:
        http://www.hackerfactor.com/blog/index.php?/archives/432-Looks-Like-It.html
    """
    assert khandy.is_numpy_image(image)
    if image.ndim == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    resized = cv2.resize(image, (8, 8))
    
    mean_val = np.mean(resized)
    hash_mat = resized >= mean_val
    hash_val = _convert_bool_matrix_to_int(hash_mat)
    return f'{hash_val:016x}'
    
    
def calc_image_dhash(image):
    """Difference Hashing

    References:
        http://www.hackerfactor.com/blog/index.php?/archives/432-Looks-Like-It.html
    """
    assert khandy.is_numpy_image(image)
    if image.ndim == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    resized = cv2.resize(image, (9, 8))
    
    hash_mat = resized[:,:-1] >= resized[:,1:]
    hash_val = _convert_bool_matrix_to_int(hash_mat)
    return f'{hash_val:016x}'
    
    
def calc_image_phash(image):
    """Perceptual Hashing
    
    References:
        http://www.hackerfactor.com/blog/index.php?/archives/432-Looks-Like-It.html
    """
    assert khandy.is_numpy_image(image)
    if image.ndim == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    resized = cv2.resize(image, (32, 32))

    dct_coeff = cv2.dct(resized.astype(np.float32))
    reduced_dct_coeff = dct_coeff[:8, :8]

    # # mean of coefficients excluding the DC term (0th term)
    # mean_val = np.mean(reduced_dct_coeff.flatten()[1:])
    # median of coefficients
    median_val = np.median(reduced_dct_coeff)

    hash_mat = reduced_dct_coeff >= median_val
    hash_val = _convert_bool_matrix_to_int(hash_mat)
    return f'{hash_val:016x}'
    


import copy
import itertools
import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union, Callable

import numpy as np

import khandy

torch = khandy.import_torch()

__all__ = ['DetObjectItem', 'DetObjectSortDir', 'DetObjectSortBy', 'DetObjects', 
           'BaseDetector', 'convert_det_objects_to_detect_ir_record', 
           'convert_detect_ir_record_to_det_objects',
           'convert_det_objects_to_detect_ir', 'convert_detect_ir_to_det_objects',
           'concat_det_objects', 'detect_in_det_objects']


@dataclass
class DetObjectItem:
    x_min: float
    y_min: float
    x_max: float
    y_max: float
    conf: float
    class_index: int
    class_name: str
    _extra_fields: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        for name, value in self._extra_fields.items():
            assert name not in self.__annotations__, f'Extra field {name} conflicts with existing field'
            assert len(value) == 1, f'Extra field {name} must have length 1, got {len(value)}'
            
    def __getattr__(self, name: str) -> Any:
        try:
            return self._extra_fields[name]
        except KeyError:
            # NB: use super().__getattribute__ instead of super().__getattr__
            # type object 'object' has no attribute '__getattr__'
            return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name not in self.__annotations__ and name not in ['area']:
            assert len(value) == 1, f'Extra field {name} must have length 1, got {len(value)}'
            self._extra_fields[name] = value
        else:
            super().__setattr__(name, value)
            
    @property
    def area(self) -> float:
        return (self.x_max - self.x_min) * (self.y_max - self.y_min)

    def to_det_objects(self) -> "DetObjects":
        box = [self.x_min, self.y_min, self.x_max, self.y_max]
        kwargs = {
            'boxes': [box],
            'confs': [self.conf],
            'classes': [self.class_index],
            'class_names': [self.class_name],
            **self._extra_fields
        }
        return DetObjects(**kwargs)
    
    
class DetObjectSortDir(Enum):
    ASC = auto()
    DESC = auto()


class DetObjectSortBy(Enum):
    BY_AREA = auto()
    BY_CONF = auto()
    BY_CLASS = auto()


class DetObjects(khandy.EqLenSequences):
    boxes: khandy.KArray
    confs: khandy.KArray
    classes: khandy.KArray
    class_names: List[str]

    def __init__(
        self,
        boxes: Optional[khandy.KArray] = None,
        confs: Optional[khandy.KArray] = None,
        classes: Optional[khandy.KArray] = None,
        class_names: Optional[List[str]] = None,
        **kwargs
    ) -> None:
        super().__init__(
            boxes=boxes, 
            confs=confs, 
            classes=classes, 
            class_names=class_names, 
            **kwargs
        )
    
    def __setattr__(self, name: str, value: Any) -> None:
        if name == 'boxes':
            value = self._set_boxes(value)
        elif name == 'confs':
            value = self._set_confs(value)
        elif name == 'classes':
            value = self._set_classes(value)
        elif name == 'class_names':
            value = self._set_class_names(value)
        super().__setattr__(name, value)
        
    def _set_boxes(self, boxes: Optional[khandy.KArray] = None) -> khandy.KArray:
        if boxes is None:
            boxes = np.empty((len(self), 4), dtype=np.float32)
        if torch is not None and isinstance(boxes, torch.Tensor):
            pass
        elif isinstance(boxes, np.ndarray):
            pass
        else:
            if len(boxes) == 0:
                boxes = np.empty((0, 4), dtype=np.float32)
            else:
                boxes = np.asarray(boxes, dtype=np.float32)

        assert boxes.ndim == 2, f'boxes ndim is not 2, got {boxes.ndim}'
        assert boxes.shape[1] == 4, f'boxes last axis size is not 4, got {boxes.shape[1]}'
        return boxes
 
    def _set_confs(self, confs: Optional[khandy.KArray] = None) -> khandy.KArray:
        if isinstance(self.boxes, np.ndarray):
            if confs is None:
                confs = np.ones((len(self),), dtype=np.float32)
            else:
                confs = np.asarray(confs, dtype=np.float32)
        else:
            if confs is None:
                confs = torch.ones((len(self),), dtype=torch.float32, device=self.boxes.device)
            else:
                confs = torch.as_tensor(confs, dtype=torch.float32, device=self.boxes.device)

        if confs.ndim == 2 and confs.shape[-1] == 1:
            confs = confs.squeeze(1)
        assert confs.ndim == 1, f'confs ndim is not 1, got {confs.ndim}'
        return confs

    def _set_classes(self, classes: Optional[khandy.KArray] = None) -> khandy.KArray:
        if isinstance(self.boxes, np.ndarray):
            if classes is None:
                classes = np.zeros((len(self),), dtype=np.int32)
            else:
                classes = np.asarray(classes, dtype=np.int32)
        else:
            if classes is None:
                classes = torch.zeros((len(self),), dtype=torch.int32, device=self.boxes.device)
            else:
                classes = torch.as_tensor(classes, dtype=torch.int32, device=self.boxes.device)

        if classes.ndim == 2 and classes.shape[-1] == 1:
            classes = classes.squeeze(1)
        assert classes.ndim == 1, f'classes ndim is not 1, got {classes.ndim}'
        return classes

    def _set_class_names(self, class_names: Optional[List[str]] = None) -> List[str]:
        if class_names is None:
            class_names = [f'unnamed_class#{class_ind}' for class_ind in self.classes]
            
        assert khandy.is_seq_of(class_names, str), f'class_names must be list of str'
        return class_names
    
    def __getitem__(self, key: Union[int, slice]) -> Union["DetObjects", DetObjectItem]:
        item = super().__getitem__(key)
        if type(key) == int:
            _extra_fields = {name: getattr(self, name) for name in self._fields 
                            if name not in ['boxes', 'confs', 'classes', 'class_names']}
            return DetObjectItem(
                x_min=item.boxes[0, 0].item(),
                y_min=item.boxes[0, 1].item(),
                x_max=item.boxes[0, 2].item(),
                y_max=item.boxes[0, 3].item(),
                conf=item.confs[0].item(),
                class_index=item.classes[0].item(),
                class_name=item.class_names[0],
                _extra_fields=_extra_fields
            )
        return item
    
    def filter_by_class_index(self, interested_class_inds, inplace=False) -> "DetObjects":
        warnings.warn(
            "filter_by_class_index is deprecated, use filter_by_class_indices instead.",
            DeprecationWarning
        )
        assert isinstance(self.classes, np.ndarray)
        mask = np.zeros((len(self.classes),), dtype=bool)
        for class_ind in interested_class_inds:
            mask = np.logical_or(mask, self.classes == class_ind)
        keep = np.nonzero(mask)[0]
        return self.filter(keep, inplace)

    def filter_by_class_indices(
        self,
        interested: Optional[Union[Tuple[int, ...], List[int]]] = None,
        ignored: Optional[Union[Tuple[int, ...], List[int]]] = None,
        inplace: bool = False
    ) -> "DetObjects":
        if interested is not None and ignored is not None:
            raise ValueError("You cannot specify both 'interested' and 'ignored' at the same time.")
        if interested is None and ignored is None:
            return self if inplace else copy.deepcopy(self)
        
        if ignored is not None:
            mask = np.isin(self.classes, ignored, invert=True)
        elif interested is not None:
            mask = np.isin(self.classes, interested)

        keep = np.nonzero(mask)[0]
        return self.filter(keep, inplace)
    
    def filter_by_class_names(
        self, 
        interested: Optional[Union[Tuple[str, ...], List[str]]] = None,
        ignored: Optional[Union[Tuple[str, ...], List[str]]] = None, 
        inplace: bool = False
    ) -> "DetObjects":
        if interested is not None and ignored is not None:
            raise ValueError("You cannot specify both 'interested' and 'ignored' at the same time.")
        if interested is None and ignored is None:
            return self if inplace else copy.deepcopy(self)
        
        if ignored is not None:
            mask = np.isin(self.class_names, ignored, invert=True)
        elif interested is not None:
            mask = np.isin(self.class_names, interested)

        keep = np.nonzero(mask)[0]
        return self.filter(keep, inplace)

    def filter_by_area(
        self, 
        min_area: Union[int, float], 
        inplace: bool = False
    ) -> "DetObjects":
        assert isinstance(self.boxes, np.ndarray)
        widths = self.boxes[:, 2] - self.boxes[:, 0]
        heights = self.boxes[:, 3] - self.boxes[:, 1] 
        mask = widths * heights >= min_area
        keep = np.nonzero(mask)[0]
        return self.filter(keep, inplace)
    
    def filter_by_min_area(
        self, 
        min_area: Union[int, float], 
        inplace: bool = False
    ) -> "DetObjects":
        warnings.warn(
            "filter_by_min_area is deprecated, use filter_by_area instead.",
            DeprecationWarning
        )
        return self.filter_by_area(min_area, inplace)

    def filter_by_size(
        self, 
        min_width: Optional[Union[int, float]] = None, 
        min_height: Optional[Union[int, float]] = None,
        inplace: bool = False
    ) -> "DetObjects":
        assert isinstance(self.boxes, np.ndarray)
        keep = khandy.filter_small_boxes(self.boxes, min_width, min_height)
        return self.filter(keep, inplace)

    def filter_by_min_size(
        self, 
        min_width: Optional[Union[int, float]] = None, 
        min_height: Optional[Union[int, float]] = None,
        inplace: bool = False
    ) -> "DetObjects":
        warnings.warn(
            "filter_by_min_size is deprecated, use filter_by_size instead.",
            DeprecationWarning
        )
        return self.filter_by_size(min_width, min_height, inplace)
    
    def filter_by_conf(
        self, 
        conf_thresh: Optional[Union[float, List[float], Tuple[float], np.ndarray]], 
        inplace: bool = False
    ) -> "DetObjects":
        assert isinstance(self.confs, np.ndarray)
        if isinstance(conf_thresh, (list, tuple, np.ndarray)):
            conf_thresh = np.array(conf_thresh)
            mask = self.confs > conf_thresh[self.classes]
        else:
            mask = self.confs > conf_thresh
        keep = np.nonzero(mask)[0]
        return self.filter(keep, inplace)

    def filter_by_func(
        self, 
        func: Callable[[DetObjectItem], bool], 
        inplace: bool = False
    ) -> "DetObjects":
        assert isinstance(self.boxes, np.ndarray)
        mask = np.array([func(det_object) for det_object in self], dtype=bool)
        keep = np.nonzero(mask)[0]
        return self.filter(keep, inplace)

    def nms(self, iou_thresh, ratio_type='iou', inplace=False) -> "DetObjects":
        assert isinstance(self.confs, np.ndarray)
        keep = khandy.non_max_suppression(self.boxes, self.confs, iou_thresh, self.classes, ratio_type)
        return self.filter(keep, inplace)

    def sort(self, sort_by: DetObjectSortBy, direction:  DetObjectSortDir = DetObjectSortDir.DESC) -> "DetObjects":
        assert isinstance(self.confs, np.ndarray)
        if sort_by == DetObjectSortBy.BY_CONF:
            sorted_inds = np.argsort(self.confs, axis=0)
        elif sort_by == DetObjectSortBy.BY_AREA:
            boxes = khandy.Boxes(self.boxes)
            sorted_inds = np.argsort(boxes.areas, axis=0)
        elif sort_by == DetObjectSortBy.BY_CLASS:
            sorted_inds = np.argsort(self.classes, axis=0)
        if direction == DetObjectSortDir.DESC:
            sorted_inds = sorted_inds[::-1]
        return self[sorted_inds]


class BaseDetector(ABC):
    def __init__(
        self, 
        num_classes: Optional[int] = None,
        conf_thresh: Optional[Union[float, List[float], Tuple[float], np.ndarray]] = None,
        iou_thresh: Optional[float] = None,
        min_width: Optional[Union[int, float]] = None,
        min_height: Optional[Union[int, float]] = None,
        min_area: Optional[Union[int, float]] = None,
        class_names: Optional[Union[List[str], Tuple[str]]] = None,
        sort_by: Optional[DetObjectSortBy] = None,
        sort_dir: Optional[DetObjectSortDir] = DetObjectSortDir.DESC
    ):
        self._num_classes = num_classes
        self._conf_thresh = conf_thresh
        self._iou_thresh = iou_thresh
        self._min_width = min_width
        self._min_height = min_height
        self._min_area = min_area
        self._class_names = class_names
        self._sort_by = sort_by
        self._sort_dir = sort_dir

    @property
    def num_classes(self) -> Optional[int]:
        return self._num_classes
    
    @property
    def conf_thresh(self) -> Optional[Union[float, np.ndarray]]:
        return self._conf_thresh
    
    @conf_thresh.setter
    def conf_thresh(self, value: Optional[Union[float, List[float], Tuple[float], np.ndarray]]):
        if value is None or isinstance(value, float):
            pass
        elif isinstance(value, (list, tuple)):
            assert khandy.is_seq_of(value, float) and len(value) == self.num_classes
            value = np.array(value)
        elif isinstance(value, np.ndarray):
            assert value.shape == (self.num_classes,) or value.shape == (self.num_classes, 1)
            value = value.flatten()
        else:
            raise TypeError(f'Unsupported type for conf_thresh, got {type(value)}')
        self._conf_thresh = value
    
    @property
    def iou_thresh(self) -> Optional[float]:
        return self._iou_thresh
    
    @iou_thresh.setter
    def iou_thresh(self, value: Optional[float]):
        assert value is None or isinstance(value, float), f'Unsupported type for iou_thresh, got {type(value)}'
        assert value is None or (0 < value < 1), f'iou_thresh must be in (0, 1), got {value}'
        self._iou_thresh = value
        
    @property
    def min_width(self) -> Optional[Union[int, float]]:
        return self._min_width
    
    @min_width.setter
    def min_width(self, value: Optional[Union[int, float]]):
        assert value is None or isinstance(value, (int, float)), f'Unsupported type for min_width, got {type(value)}'
        assert value is None or value >= 0, f'min_width must be >= 0, got {value}'
        self._min_width = value
        
    @property
    def min_height(self) -> Optional[Union[int, float]]:
        return self._min_height
    
    @min_height.setter
    def min_height(self, value: Optional[Union[int, float]]):
        assert value is None or isinstance(value, (int, float)), f'Unsupported type for min_height, got {type(value)}'
        assert value is None or value >= 0, f'min_height must be >= 0, got {value}'
        self._min_height = value
        
    @property
    def min_area(self) -> Optional[Union[int, float]]:
        return self._min_area
    
    @min_area.setter
    def min_area(self, value: Optional[Union[int, float]]):
        assert value is None or isinstance(value, (int, float)), f'Unsupported type for min_area, got {type(value)}'
        assert value is None or value >= 0, f'min_area must be >= 0, got {value}'
        self._min_area = value
        
    @property
    def class_names(self) -> Optional[List[str]]:
        return self._class_names
    
    @class_names.setter
    def class_names(self, value: Optional[Union[List[str], Tuple[str]]]):
        if value is not None:
            assert khandy.is_seq_of(value, str) and len(value) == self.num_classes
            value = list(value)
        self._class_names = value
        
    @property
    def sort_by(self) -> Optional[DetObjectSortBy]:
        return self._sort_by
    
    @sort_by.setter
    def sort_by(self, value: Optional[DetObjectSortBy]):
        self._sort_by = value
        
    @property
    def sort_dir(self) -> DetObjectSortDir:
        return self._sort_dir
    
    @sort_dir.setter
    def sort_dir(self, value: Optional[DetObjectSortDir]):
        self._sort_dir = value or DetObjectSortDir.DESC

    @abstractmethod
    def forward(self, image: khandy.KArray, **kwargs) -> DetObjects:
        pass
    
    def filter_by_conf(self, det_objects: DetObjects) -> DetObjects:
        if self.conf_thresh is not None:
            return det_objects.filter_by_conf(self.conf_thresh, inplace=True)
        return det_objects
    
    def filter_by_size(self, det_objects: DetObjects) -> DetObjects:
        if self.min_width is not None or self.min_height is not None:
            return det_objects.filter_by_size(self.min_width, self.min_height, inplace=True)
        return det_objects
    
    def filter_by_area(self, det_objects: DetObjects) -> DetObjects:
        if self.min_area is not None:
            return det_objects.filter_by_area(self.min_area, inplace=True)
        return det_objects
    
    def nms(self, det_objects: DetObjects, ratio_type: str = 'iou') -> DetObjects:
        if self.iou_thresh is not None:
            return det_objects.nms(self.iou_thresh, ratio_type, inplace=True)
        return det_objects
    
    def __call__(self, image: khandy.KArray, **kwargs) -> DetObjects:
        det_objects = self.forward(image, **kwargs)
        if self.class_names is not None:
            det_objects.class_names = [self.class_names[ind.item()] for ind in det_objects.classes]
        if self.sort_by is not None:
            det_objects = det_objects.sort(self.sort_by, self.sort_dir, inplace=True)
        return det_objects
    

def convert_det_objects_to_detect_ir_record(
    det_objects: DetObjects, 
    image_width: int, 
    image_height: int,
    filename: str = ''
) -> khandy.label.DetectIrRecord:
    warnings.warn(
        "convert_det_objects_to_detect_ir_record is deprecated, use convert_det_objects_to_detect_ir instead.",
        DeprecationWarning
    )
    ir_record = khandy.label.DetectIrRecord(
        filename=filename, 
        width=image_width, 
        height=image_height
    )
    for det_object in det_objects:
        ir_object = khandy.label.DetectIrObject(
            label=det_object.class_name, 
            x_min=det_object.x_min, 
            y_min=det_object.y_min, 
            x_max=det_object.x_max, 
            y_max=det_object.y_max)
        ir_record.objects.append(ir_object)
    return ir_record


def convert_detect_ir_record_to_det_objects(
    ir_record: khandy.label.DetectIrRecord,
    label2index: Optional[Mapping[str, int]] = None
) -> DetObjects:
    warnings.warn(
        "convert_detect_ir_record_to_det_objects is deprecated, use convert_detect_ir_to_det_objects instead.",
        DeprecationWarning
    )
    class_names, boxes = [], []
    for ir_object in ir_record.objects:
        class_names.append(ir_object.label)
        boxes.append([ir_object.x_min, ir_object.y_min, 
                      ir_object.x_max, ir_object.y_max])
    if label2index is not None:
        class_indices = [label2index[name] for name in class_names]
    else:
        class_indices = [-1 for _ in class_names]
    det_objects = DetObjects(
        boxes=boxes, 
        class_names=class_names,
        class_indices=class_indices
    )
    return det_objects


def convert_det_objects_to_detect_ir(
    det_objects: DetObjects, 
) -> List[khandy.label.DetectIrObject]:
    ir_objects = []
    for det_object in det_objects:
        ir_object = khandy.label.DetectIrObject(
            label=det_object.class_name, 
            x_min=det_object.x_min, 
            y_min=det_object.y_min, 
            x_max=det_object.x_max, 
            y_max=det_object.y_max)
        ir_objects.append(ir_object)
    return ir_objects


def convert_detect_ir_to_det_objects(
    detect_ir: Union[khandy.label.DetectIrRecord, 
                     khandy.label.DetectIrObject, 
                     List[khandy.label.DetectIrObject]],
    label2index: Optional[Mapping[str, int]] = None
) -> DetObjects:
    if isinstance(detect_ir, khandy.label.DetectIrRecord):
        ir_objects = detect_ir.objects
    elif isinstance(detect_ir, khandy.label.DetectIrObject):
        ir_objects = [detect_ir]
    else:
        ir_objects = detect_ir
        
    class_names, boxes = [], []
    for ir_object in ir_objects:
        class_names.append(ir_object.label)
        boxes.append([ir_object.x_min, ir_object.y_min, 
                      ir_object.x_max, ir_object.y_max])
    if label2index is not None:
        class_indices = [label2index[name] for name in class_names]
    else:
        class_indices = [-1 for _ in class_names]
    det_objects = DetObjects(
        boxes=boxes, 
        class_names=class_names,
        class_indices=class_indices
    )
    return det_objects


def _concatenate_arrays_or_sequences(
    arrays_or_sequences: Union[List[khandy.KArray], List[Sequence]]
) -> Union[khandy.KArray, Sequence]:
    assert len(arrays_or_sequences) > 0
    if khandy.is_torch_available() and khandy.is_list_of(arrays_or_sequences, torch.Tensor):
        return torch.cat(arrays_or_sequences, dim=0)
    elif khandy.is_list_of(arrays_or_sequences, np.ndarray):
        return np.concatenate(arrays_or_sequences, axis=0)
    elif khandy.is_list_of(arrays_or_sequences, Sequence):
        first_type = type(arrays_or_sequences[0])
        return first_type(itertools.chain.from_iterable(arrays_or_sequences))
    else:
        raise TypeError('Unsupported type!')
    
    
def concat_det_objects(det_objects_list: List[DetObjects], only_common_fields: bool = False) -> DetObjects:
    """Concatenates a list of DetObjects into a single DetObjects instance.
    Args:
        det_objects_list (List[DetObjects]): A list of DetObjects instances to concatenate.
        only_common_fields (bool, optional): If True, only the fields common to all DetObjects will be concatenated.
            If False, all DetObjects must have exactly the same fields. Defaults to False.

    Returns:
        DetObjects: A new DetObjects instance with concatenated fields.

    Raises:
        ValueError: If only_common_fields is False and the DetObjects do not have identical fields.
    """
    if len(det_objects_list) == 0:
        common_fields = set()
    else:
        all_fields = [set(det_objects.get_fields()) for det_objects in det_objects_list]
        if only_common_fields:
            common_fields = set.intersection(*all_fields)
        else:
            if not all(fields == all_fields[0] for fields in all_fields):
                raise ValueError("All DetObjects must have the same fields when only_common_fields is False.")
            common_fields = all_fields[0]
    
    name_to_list = {}
    for det_objects in det_objects_list:
        for name in common_fields:
            name_to_list.setdefault(name, []).append(getattr(det_objects, name))
    name_to_sequence = {}
    for name, values in name_to_list.items():
        name_to_sequence[name] = _concatenate_arrays_or_sequences(values)
    return DetObjects(**name_to_sequence)


def detect_in_det_objects(
    detector: BaseDetector, 
    image: np.ndarray, 
    det_objects: DetObjects, 
    min_area: Optional[Union[int, float]] = None,
    **detector_kwargs
) -> DetObjects:
    dst_det_objects_list = []
    for det_object in det_objects:
        if min_area is not None and det_object.area < min_area:
            continue
        
        x_min = round(det_object.x_min)
        y_min = round(det_object.y_min)
        x_max = round(det_object.x_max)
        y_max = round(det_object.y_max)
        cropped = khandy.crop_image(image, x_min, y_min, x_max, y_max)
        objects_in_object = detector(cropped, **detector_kwargs)
        objects_in_object.boxes[:, 0] += det_object.x_min
        objects_in_object.boxes[:, 1] += det_object.y_min
        objects_in_object.boxes[:, 2] += det_object.x_min
        objects_in_object.boxes[:, 3] += det_object.y_min
        dst_det_objects_list.append(objects_in_object)
    return concat_det_objects(dst_det_objects_list)

